# -*- coding: utf-8 -*-
"""
Test the :mod:`xoa.coords` module
"""

import pytest
import numpy as np
import xarray as xr

import xoa
from xoa import coords


@pytest.mark.parametrize(
    "inshape,indims,tdims,mode,outshape,outdims",
    [
     ((1, 2), ('y', 'x'), ("x", "y"), "classic",
      (2, 1), ("x", "y")),
     ((3, 2), ('y', 'x'), ("t", "y"), "insert",
      (2, 1, 3), ("x", "t", "y")),
     ((3, 2), ('y', 'x'), ("x", "t", "y"), "compat",
      (2, 3), ("x", "y")),
     ((3, 4, 2), ('y', 't', 'x'), (Ellipsis, "x", "e", "y"), "compat",
      (4, 2, 3), ("t", "x", "y")),
    ]
)
def test_coords_transpose(inshape, indims, tdims, mode, outshape, outdims):
    da = xr.DataArray(np.ones(inshape), dims=indims)
    dao = coords.transpose(da, tdims, mode)
    assert dao.dims == outdims
    assert dao.shape == outshape


def test_coords_is_lon():

    x = xr.DataArray([5], dims="lon", name="lon")
    y = xr.DataArray([5], dims="lat")
    temp = xr.DataArray(np.ones((1, 1)), dims=('lat', 'lon'), coords={'lon': x, 'lat': y})

    assert coords.is_lon(x)
    assert coords.is_lon(temp.lon)
    assert not coords.is_lon(temp.lat)


def test_coords_get_depth_from_variable():

    da = xr.DataArray(
        np.ones((2, 3)),
        dims=("depth", "lon"),
        coords={"depth": ("depth", [0, 1]), "lon": [1, 2, 3]})
    depth = coords.get_depth(da)
    assert depth is not None
    np.testing.assert_allclose(depth.values, [0, 1])


def test_coords_get_depth_from_sigma():

    ds = xoa.open_data_sample("croco.south-africa.meridional.nc")
    depth = coords.get_depth(ds)
    assert depth is not None
    assert depth.name == "depth"


def test_coords_get_depth_from_dz():

    ds = xoa.open_data_sample("hycom.gdp.h.nc")
    ds = ds.rename(h="dz")
    depth = coords.get_depth(ds)
    assert depth is not None
    assert depth.name == "depth"



# -*- coding: utf-8 -*-
"""
Test the :mod:`xoa.sigma` module
"""

import xoa
from xoa import sigma


def test_sigma_decode_cf_sigma():

    ds = xoa.open_data_sample("croco.south-africa.meridional.nc")
    dsd = sigma.decode_cf_sigma(ds)

    assert "depth" in dsd
    assert dsd.depth.dims == ('time', 's_rho', 'eta_rho', 'xi_rho')
    assert dsd.depth.shape == (1, 32, 56, 1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Geographic utilities
"""
# Copyright 2020-2021 Shom
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import numpy as np
import numba

from . import coords as xcoords


#: Earth radius in meters
EARTH_RADIUS = 6371e3


@numba.vectorize
def _haversine_(lon0, lat0, lon1, lat1):
    """Haversine distance between two points on a **unit sphere**

    Parameters
    ----------
    lon0: float, array_like
        Longitude of the first point(s)
    lat0: float, array_like
        Latitude of the first point(s)
    lon1: float, array_like
        Longitude of the second point(s)
    lat1: float, array_like
        Latitude of the second point(s)

    Return
    ------
    float, array_like
        Distance(s)
    """
    deg2rad = math.pi / 180.
    dist = math.sin(deg2rad*(lat0-lat1)*0.5)**2
    dist += (math.cos(deg2rad*lat0) * math.cos(deg2rad*lat1) *
             math.sin(deg2rad*(lon0-lon1)*0.5)**2)
    dist = 2. * math.asin(math.sqrt(dist))
    return dist


def haversine(lon0, lat0, lon1, lat1, radius=EARTH_RADIUS):
    """Haversine distance between two points

    Parameters
    ----------
    lon0: float, array_like
        Longitude of the first point(s)
    lat0: float, array_like
        Latitude of the first point(s)
    lon1: float, array_like
        Longitude of the second point(s)
    lat1: float, array_like
        Latitude of the second point(s)
    radius: float
        Radius of the sphere which defaults to the earth radius

    Return
    ------
    float, array_like
        Distance(s)
    """
    return _haversine_(
        np.double(lon0), np.double(lat0),
        np.double(lon1), np.double(lat1)) * radius


def bearing(lon0, lat0, lon1, lat1):
    """Compute the bearing angle (forward azimuth)

    Parameters
    ----------
    lon0: float, array_like
        Longitude of the first point(s)
    lat0: float, array_like
        Latitude of the first point(s)
    lon1: float, array_like
        Longitude of the second point(s)
    lat1: float, array_like
        Latitude of the second point(s)

    Return
    ------
    float, array_like
        Angle(s)
    """
    return _bearing_(
        np.double(lon0), np.double(lat0), np.double(lon1), np.double(lat1))


@numba.vectorize
def _bearing_(lon0, lat0, lon1, lat1):
    deg2rad = math.pi / 180.
    a = math.arctan2(
        math.cos(deg2rad*lat0)*math.sin(deg2rad*lat1) -
        math.sin(deg2rad*lat0)*math.cos(deg2rad*lat1)*math.cos(deg2rad*(lon1-lon0)),
        math.sin(deg2rad*(lon1-lon0))*math.cos(deg2rad*lat1))
    return a * 180 / math.pi


def get_extent(extent, margin=0, square=False):
    """Compute the geographic extent in degrees

    Parameters
    ----------
    extent: xarray.DataArray, xarray.Dataset, dict, tuple, list
        Either:

        - An array or dataset with longitude and latitude coordinates.
        - A dict with ``lon`` and ``lat`` keys: ``dict(lon=..., lat=...)``
        - A two-element tuple of longitudes and latitudes: ``(lon, lat)``
        - A extent list: ``[xmin, xmax, ymin, ymax]``.
    margin: float
        A relative fraction of the width and height that is used to set margins.
        For instance, a value of ``-0.1`` shrinks the box of 10% on each side.
    square: bool
        Force the box to be square in degrees.

    Return
    ------
    list
        ``[xmin, xmax, ymin, ymax]``

    Example
    -------

    .. ipython:: python

        @suppress
        from xoa.geo import get_extent
        @suppress
        import numpy as np
        get_extent([10., 20., 10., 20.], margin=0.1)
        get_extent({"lon": np.linspace(10, 20, 5), "lat": np.linspace(10, 20, 5)}, square=True)
        get_extent((np.linspace(10, 20, 5), np.linspace(10, 20, 5)), margin=-.1, square=True)

    """
    # Get min and max
    if hasattr(extent, "coords"):
        lon = xcoords.get_lon(extent)
        lat = xcoords.get_lat(extent)
        extent = (lon, lat)
    elif isinstance(extent, dict):
        extent = extent['lon'], extent['lat']
    if isinstance(extent, (list, np.ndarray)):
        xmin, xmax, ymin, ymax = extent
    else:  # tuple
        lon, lat = extent
        xmin = float(np.min(lon))
        xmax = float(np.max(lon))
        ymin = float(np.min(lat))
        ymax = float(np.max(lat))

    # Scale
    if square or margin:
        dx = xmax - xmin
        dy = ymax - ymin
        x0 = 0.5 * (xmin+xmax)
        y0 = 0.5 * (ymin+ymax)
        if square:
            aspect = dx * math.cos(y0*math.pi/180) / dy
            if aspect > 1:
                dy *= aspect
            else:
                dx /= aspect
        xmargin = margin * dx
        ymargin = margin * dy
        xmin = x0 - 0.5 * dx - xmargin
        xmax = x0 + 0.5 * dx + xmargin
        ymin = y0 - 0.5 * dy - ymargin
        ymax = y0 + 0.5 * dy + ymargin

    return [xmin, xmax, ymin, ymax]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Test the :mod:`xoa.cf` module
"""

# from unittest.mock import Mock
import warnings
import pytest

import numpy as np
import xarray as xr

import xoa
from xoa import cf


@pytest.mark.parametrize(
    "attr,value,expected",
    [
        ("standard_name", "my_var_at_t_location", ("my_var", "t")),
        ("standard_name", "my_var", ("my_var", None)),
        ("long_name", "My var at T location", ("My var", "t")),
        ("name", "myvar_t", ("myvar", "t")),
    ],
)
def test_cf_sglocator_parse_attr(attr, value, expected):
    assert cf.SGLocator().parse_attr(attr, value) == expected


@pytest.mark.parametrize(
    "attr,value,expected",
    [
        ("standard_name", "my_var_at_t_location", ("my_var_at_t_location", None)),
        ("standard_name", "my_var_at_u_location", ("my_var", "u")),
        ("long_name", "My var at RHO location", ("My var", "rho")),
        ("long_name", "My var at rho location", ("My var", "rho")),
        ("name", "myvarrho", ("myvar", "rho")),
    ],
)
def test_cf_sglocator_parse_attr_with_valid_locations(attr, value, expected):
    assert (
        cf.SGLocator(
            valid_locations=['u', 'rho'],
            name_format="{root}{loc}",
        ).parse_attr(attr, value)
        == expected
    )


@pytest.mark.parametrize(
    "name,standard_name,long_name,loc",
    [
        ("u_t", None, None, "t"),
        (None, "u_at_t_location", None, "t"),
        (None, None, "U at T location", "t"),
        ("u_t", "_at_t_location", "U at T location", "t"),
        ("u", "u_at_t_location", None, "t"),
        ("u", "u", "U", None),
    ],
)
def test_cf_sglocator_get_loc_from_da(name, standard_name, long_name, loc):

    da = xr.DataArray(0)
    if name:
        da.name = name
    if standard_name:
        da.attrs["standard_name"] = standard_name
    if long_name:
        da.attrs["long_name"] = long_name

    parsed_loc = cf.SGLocator().get_loc_from_da(da)
    assert parsed_loc == loc


@pytest.mark.parametrize(
    "name,standard_name,long_name",
    [
        ("u_t", "u_at_u_location", None),
        (None, "u_at_t_location", "U at U location"),
        ("u_u", None, "U at T location"),
        ("u_u", "u_at_w_location", "U at T location"),
    ],
)
def test_cf_sglocator_get_loc_from_da_error(name, standard_name, long_name):

    da = xr.DataArray(0)
    if name:
        da.name = name
    if standard_name:
        da.attrs["standard_name"] = standard_name
    if long_name:
        da.attrs["long_name"] = long_name

    with pytest.raises(cf.XoaCFError):
        cf.SGLocator().get_loc_from_da(da, errors="raise")

    cf.SGLocator().get_loc_from_da(da, errors="ignore")


@pytest.mark.parametrize(
    "attr,root,loc,expected",
    [
        ("standard_name", "my_var", "t", True),
        ("standard_name", "my_var2", "t", False),
        ("standard_name", "my_var", "x", False),
        ("standard_name", "my_var", None, True),
        ("standard_name", "my_var", "xtu", True),
        ("long_name", "My var", "t", True),
        ("long_name", "My var", "x", False),
        ("name", "myvar", "t", True),
        ("name", "myvar", "x", False),
    ],
)
def test_cf_sglocator_match_attr(attr, root, loc, expected):
    value = dict(
        standard_name="my_var_at_t_location",
        long_name="My var at T location",
        name="myvar_t",
    )[attr]
    assert cf.SGLocator().match_attr(attr, value, root, loc) is expected


@pytest.mark.parametrize(
    "attr,root,loc,expected",
    [
        ("standard_name", "my_var", "t", "my_var_at_t_location"),
        ("standard_name", "my_var", "", "my_var"),
        ("long_name", "My var", "t", "My var at T location"),
        ("name", "myvar", "t", "myvar_t"),
    ],
)
def test_cf_sglocator_format_attr(attr, root, loc, expected):
    assert cf.SGLocator().format_attr(attr, root, loc) == expected


def test_cf_sglocator_format_attr_valid_locations():
    with pytest.raises(cf.XoaCFError) as excinfo:
        cf.SGLocator(valid_locations="x").format_attr("name", "banana", "y")
    assert str(excinfo.value) == (
        'Location "y" is not recognised by the currents '
        'specifications. Registered locations are: x'
    )


def test_cf_sglocator_format_attrs_no_loc():
    attrs = {
        "name": "u_u",
        "standard_name": "banana_at_t_location",
        "long_name": "Banana at T location",
        "int_attr": 10,
        "str_attr": "good",
    }

    fmt_attrs = cf.SGLocator().format_attrs(attrs, loc='')
    assert fmt_attrs["name"] == "u_u"
    assert fmt_attrs["standard_name"] == "banana"
    assert fmt_attrs["long_name"] == "Banana"
    for attr in ("int_attr", "str_attr"):
        assert fmt_attrs[attr] == attrs[attr]


def test_cf_sglocator_format_attrs_with_loc():
    attrs = {
        "name": "u_u",
        "standard_name": "banana_at_t_location",
        "long_name": "Banana",
        "int_attr": 10,
        "str_attr": "good",
    }

    fmt_attrs = cf.SGLocator().format_attrs(attrs, loc="f")
    assert fmt_attrs["name"] == "u_u"
    assert fmt_attrs["standard_name"] == "banana_at_f_location"
    assert fmt_attrs["long_name"] == "Banana at F location"
    for attr in ("int_attr", "str_attr"):
        assert fmt_attrs[attr] == attrs[attr]


@pytest.mark.parametrize(
    "value0, value1, loc, value",
    [
        ("sst", None, "t", "sst_t"),
        (None, "sst", "t", "sst_t"),
        ("sst", "sss", "t", "sss_t"),
        ("sst_x", "sss_y", "t", "sss_t"),
        ("sst_t", None, None, "sst_t"),
        (None, "sst_t", None, "sst_t"),
        ("sst_x", "sss_y", None, "sss_y"),
        ("sst_x", "sss", None, "sss_x"),
        ("sst", "sss_y", None, "sss_y"),
    ],
)
def test_cf_sglocator_merge_attr(value0, value1, loc, value):
    out = cf.SGLocator().merge_attr("name", value0, value1, loc)
    assert out == value


@pytest.mark.parametrize(
    "isn, psn, osn, loc, replace",
    [
        ("sst", None, "sst", None, False),
        (None, "sst", "sst", None, False),
        ("sst", "temp", "sst", None, False),
        ("sst", "temp", "temp", None, True),
        ("sst_at_t_location", "temp_at_u_location", "sst_at_t_location", None, False),
        ("sst_at_t_location", "temp_at_u_location", "temp_at_u_location", None, True),
        ("sst", "temp", "sst_at_u_location", "u", False),
        ("sst", "temp", "temp_at_u_location", "u", True),
        ("sst_at_t_location", "temp_at_x_location", "sst_at_u_location", "u", False),
        ("sst_at_t_location", "temp_at_x_location", "temp_at_u_location", "u", True),
    ],
)
def test_cf_sglocator_patch_attrs(isn, psn, osn, loc, replace):

    iattrs = {"units": "m", "color": "blue"}
    patch = {"cmap": "viridis", "mylist": [1, 2], "units": "cm"}
    if isn:
        iattrs["standard_name"] = isn
    if psn:
        patch["standard_name"] = psn

    oattrs = cf.SGLocator().patch_attrs(iattrs, patch, loc=loc, replace=replace)

    assert oattrs["units"] == ("cm" if replace else "m")
    assert oattrs["color"] == "blue"
    assert oattrs["cmap"] == "viridis"
    assert oattrs["mylist"] == [1, 2]

    assert oattrs.get("standard_name") == osn


@pytest.mark.parametrize(
    "floc,fname,fattrs,out_name,out_standard_name,replace_attrs",
    [
        # ("p", None, None, "banana_p", "banana_at_p_location", False),
        # (None, None, None, "banana_t", "banana", False),

        # ("p", "sst", {"standard_name": "potatoe"}, "sst_p", "banana_at_p_location", False),
        # ("p", "sst", {"standard_name": "potatoe"}, "sst_p", "potatoe_at_p_location", True),
        # (
        #     'x',
        #     "sst",
        #     {"standard_name": ["potatoe", "banana"]},
        #     "sst_x",
        #     "banana_at_x_location",
        #     True,
        # ),
        (None, "sst_q", {"standard_name": ["potatoe"]}, "sst_q", "potatoe_at_q_location", True),
        # (None, "sst", {"standard_name": ["potatoe"]}, "sst_t", "potatoe_at_t_location", True),
    ],
)
def test_cf_sglocator_format_dataarray(
    floc, fname, fattrs, out_name, out_standard_name, replace_attrs
):

    lon = xr.DataArray(range(5), dims="lon")
    banana = xr.DataArray(
        lon + 20,
        dims="lon",
        coords=[lon],
        name="banana_t",
        attrs={"standard_name": "banana", "taste": "good"},
    )
    banana_fmt = cf.SGLocator().format_dataarray(
        banana, loc=floc, name=fname, attrs=fattrs, replace_attrs=replace_attrs
    )
    assert banana_fmt.name == out_name
    assert banana_fmt.standard_name == out_standard_name
    assert banana_fmt.taste == "good"


def test_cf_sglocator_format_dataarray_no_copy_no_rename():
    banana = xr.DataArray(1, name="banana_t", attrs={"standard_name": "banana"})
    banana_fmt = cf.SGLocator().format_dataarray(banana, "p", copy=False, rename=False)
    assert banana_fmt is banana
    assert banana_fmt.name == "banana_t"
    assert banana_fmt.standard_name == "banana_at_p_location"


@pytest.mark.parametrize("cache", ["ignore", "write", "rw", "read", "ignore", "clean", "rw"])
def test_cf_get_cfg_specs(cache):
    assert isinstance(cf.get_cf_specs(cache=cache), cf.CFSpecs)


def test_cf_get_cfg_specs_var():
    specs = cf.get_cf_specs().data_vars["temp"]
    assert specs["alt_names"][0] == "temperature"
    assert specs["attrs"]["standard_name"][0] == "sea_water_temperature"
    assert specs["cmap"] == "cmo.thermal"
    new_specs = cf.get_cf_specs()["temp"]
    assert new_specs is specs


def test_cf_get_cfg_specs_var_inherit():
    specs = cf.get_cf_specs().data_vars["sst"]
    assert specs["attrs"]["standard_name"][0] == "sea_surface_temperature"
    assert specs["attrs"]["units"][0] == "degrees_celsius"


def test_cf_get_cfg_specs_coord():
    specs = cf.get_cf_specs().coords["lon"]
    assert specs["alt_names"][0] == "longitude"
    assert "longitude" in specs["alt_names"]
    new_specs = cf.get_cf_specs()["lon"]
    assert new_specs is specs


def test_cf_get_cfg_specs_coord_inherit():
    specs = cf.get_cf_specs().coords["depth"]
    assert specs["alt_names"][0] == "dep"
    assert specs["attrs"]["long_name"][0] == "Depth"


@pytest.mark.parametrize(
    "cfg,key,name",
    [
        ({"data_vars": {"temp": {"alt_names": "mytemp"}}}, "temp", "mytemp"),
        ("[data_vars]\n[[sal]]\nalt_names=mysal", "sal", "mysal"),
    ],
)
def test_cf_cfspecs_load_cfg(cfg, key, name):
    cfspecs = cf.get_cf_specs()
    cfspecs.load_cfg(cfg)
    assert name in cfspecs["data_vars"][key]["alt_names"]


def test_cf_cfspecs_copy():
    cfspecs0 = cf.get_cf_specs()
    cfspecs1 = cfspecs0.copy()
    assert id(cfspecs0._dict) != id(cfspecs1._dict)
    assert sorted(list(cfspecs0._dict["data_vars"])) == sorted(list(cfspecs1._dict["data_vars"]))
    assert cfspecs0._dict["coords"] == cfspecs1._dict["coords"]
    assert cfspecs0._dict["data_vars"]["temp"] == cfspecs1._dict["data_vars"]["temp"]
    assert "temp" in cfspecs1["data_vars"]
    assert "temperature" in cfspecs1["data_vars"]["temp"]["alt_names"]


def test_cf_set_cf_specs():
    cf.reset_cache(disk=False)
    cfspecs = cf.get_cf_specs()
    cf.set_cf_specs(cfspecs)
    cf_cache = cf._get_cache_()
    assert cf_cache["current"] is cfspecs
    assert cf.get_cf_specs() is cfspecs


def test_cf_set_cf_specs_context():
    cfspecs0 = cf.get_cf_specs()
    cfspecs1 = cf.CFSpecs({"data_vars": {"temp": {"alt_names": "tempouille"}}})
    assert cf.get_cf_specs() is cfspecs0
    with cf.set_cf_specs(cfspecs1) as cfspecs:
        assert cfspecs is cfspecs1
        assert cf.get_cf_specs() is cfspecs1
    assert cf.get_cf_specs() is cfspecs0


@pytest.mark.parametrize("specialize,expected", [(False, "temp"), (True, "temperature")])
def test_cf_cfspecs_get_name(specialize, expected):
    cfspecs = cf.CFSpecs({"data_vars": {"temp": {"name": "temperature"}}})
    assert cfspecs.data_vars.get_name("temp", specialize=specialize) == expected


def test_cf_cfspecs_get_attrs():
    cfspecs = cf.get_cf_specs()
    attrs = cfspecs.data_vars.get_attrs("temp", other="ok")
    assert attrs["long_name"] == "Temperature"
    assert attrs["other"] == "ok"


def test_cf_cfspecs_get_loc_mapping():

    cf_dict0 = {
        "sglocator": {
            "valid_locations": ["u", "v"],
        },
        "data_vars": {
            "u": {
                "loc": "u",
                "add_loc": False,
                "add_coords_loc": {"lon": True, "x": True},
            },
            "bathy": {"add_loc": True},
        },
    }
    cf_specs0 = cf.CFSpecs(cf_dict0)

    ds0 = xr.Dataset(
        {"u": (("time", "y", "x"), np.ones((1, 2, 3))), "bathy": (("y", "x"), np.ones((2, 3)))},
        coords={
            "lon": (("y", "x"), np.ones((2, 3))),
            "lat": (("y", "x"), np.ones((2, 3))),
            "time": ("time", [1]),
        },
    )

    locations = cf_specs0.get_loc_mapping(ds0)
    assert locations == {'u': False, 'lon': 'u', 'x': 'u', 'bathy': 'u', 'lat': None}


@pytest.mark.parametrize("cf_name", [None, "lon"])
@pytest.mark.parametrize(
    "in_name,in_attrs",
    [
        ("lon", None),
        ("xxx", {"standard_name": "longitude"}),
        ("xxx", {"standard_name": "longitude_at_t_location"}),
        ("xxx", {"units": "degree_east"}),
    ],
)
def test_cf_cfspecs_match_coord(cf_name, in_name, in_attrs):

    lon = xr.DataArray(range(5), dims=in_name, name=in_name, attrs=in_attrs)
    res = cf.get_cf_specs().match_coord(lon, cf_name)
    if cf_name is None:
        assert res == 'lon'
    else:
        assert res is True


@pytest.mark.parametrize("cf_name", ["lon", None])
@pytest.mark.parametrize(
    "in_name,in_attrs",
    [
        ("lon", None),
        ("xxx", {"standard_name": "longitude"}),
        ("xxx", {"standard_name": "longitude_at_t_location"}),
        ("xxx", {"units": "degree_east"}),
    ],
)
def test_cf_cfspecs_search_coord(cf_name, in_name, in_attrs):

    lon = xr.DataArray(range(5), dims=in_name, name=in_name, attrs=in_attrs)
    temp = xr.DataArray(range(20, 25), dims=in_name, coords={in_name: lon}, name='temp')
    res = cf.get_cf_specs().search_coord(temp, cf_name, get="cf_name")
    assert res == 'lon'


@pytest.mark.parametrize("cf_name", ["temp", None])
@pytest.mark.parametrize(
    "in_name,in_attrs",
    [
        ("temp", None),
        ("xxx", {"standard_name": "sea_water_temperature"}),
    ],
)
def test_cf_cfspecs_match_data_var(cf_name, in_name, in_attrs):

    lon = xr.DataArray(range(5), dims='lon', name='lon')
    temp = xr.DataArray(
        range(20, 25), dims='lon', coords={'lon': lon}, name=in_name, attrs=in_attrs
    )
    res = cf.get_cf_specs().match_data_var(temp, cf_name)
    if cf_name is None:
        assert res == 'temp'
    else:
        assert res is True


@pytest.mark.parametrize("cf_name", ["temp", None])
@pytest.mark.parametrize(
    "in_name,in_attrs",
    [
        ("temp", None),
        ("xxx", {"standard_name": "sea_water_temperature"}),
    ],
)
def test_cf_cfspecs_search_data_var(cf_name, in_name, in_attrs):

    lon = xr.DataArray(range(5), dims='lon', name='lon')
    temp = xr.DataArray(
        range(20, 25), dims='lon', coords={'lon': lon}, name=in_name, attrs=in_attrs
    )
    ds = temp.to_dataset()
    assert cf.get_cf_specs().search_data_var(ds, cf_name, get="cf_name") == 'temp'


def test_cf_cfspecs_cats_get_loc_arg():

    cf_dict0 = {
        "sglocator": {
            "valid_locations": ["u", "v"],
        },
        "data_vars": {
            "u": {
                "loc": "u",
                "add_loc": False,
                "add_coords_loc": {"lon": True},
            },
        },
    }
    cf_specs0 = cf.CFSpecs(cf_dict0)

    ds0 = xr.Dataset(
        {"u": (("time", "y", "x"), np.ones((1, 2, 3))), "bathy": (("y", "x"), np.ones((2, 3)))},
        coords={
            "lon": (("y", "x"), np.ones((2, 3))),
            "lat": (("y", "x"), np.ones((2, 3))),
            "time": ("time", [1]),
        },
    )

    assert cf_specs0.coords.get_loc_arg(ds0["u"]) is None
    assert cf_specs0.coords.get_loc_arg(ds0["bathy"]) is None
    assert cf_specs0.coords.get_loc_arg(ds0["lon"]) is None

    locations0 = cf_specs0.get_loc_mapping(ds0)
    assert cf_specs0.coords.get_loc_arg(ds0["lon"], locations=locations0) == "u"

    cf_dict1 = cf_dict0.copy()
    cf_dict1["data_vars"]["u"]["add_coords_loc"]["lon"] = "v"
    cf_specs1 = cf.CFSpecs(cf_dict1)
    locations1 = cf_specs1.get_loc_mapping(ds0)
    assert cf_specs1.coords.get_loc_arg(ds0["lon"], locations=locations1) == "v"


@pytest.mark.parametrize("cf_name", [None, "lon"])
@pytest.mark.parametrize(
    "in_name,in_attrs",
    [
        ("lon", None),
        ("xxx", {"standard_name": "longitude"}),
        ("xxx", {"units": "degrees_east"}),
    ],
)
def test_cf_cfspecs_cats_format_dataarray(cf_name, in_name, in_attrs):

    lon = xr.DataArray(range(5), dims=in_name, name=in_name, attrs=in_attrs)
    lon = cf.get_cf_specs().coords.format_dataarray(lon, cf_name)
    assert lon.name == "lon"
    assert lon.standard_name == "longitude"
    assert lon.long_name == "Longitude"
    assert lon.units == "degrees_east"


def test_cf_cfspecs_cats_format_dataarray_unknown():
    coord = xr.DataArray(range(5), name='foo')
    cfspecs = cf.get_cf_specs()

    coord_fmt = cfspecs.coords.format_dataarray(coord, rename=False)
    assert coord_fmt is None

    coord_fmt = cfspecs.coords.format_dataarray(coord, rename=True)
    assert coord_fmt.name == "foo"


def test_cf_cfspecs_cats_get_allowed_names():
    cfg = {"data_vars": {"banana": {"name": "bonono", "alt_names": ["binini", "bununu"]}}}
    cfspecs = cf.CFSpecs(cfg)
    assert cfspecs.data_vars.get_allowed_names("banana") == ['banana', 'bonono', 'binini', 'bununu']


def test_cf_cfspecs_format_obj_with_loc():

    cf_dict0 = {
        "sglocator": {
            "valid_locations": ["u", "v"],
        },
        "data_vars": {
            "u": {
                "loc": "u",
                "add_loc": False,
                "add_coords_loc": {"lon": True, "x": True},
            },
            "bathy": {"add_loc": True},
        },
    }
    cf_specs0 = cf.CFSpecs(cf_dict0)

    ds0 = xr.Dataset(
        {"u": (("time", "y", "x"), np.ones((1, 2, 3))), "bathy": (("y", "x"), np.ones((2, 3)))},
        coords={
            "lon": (("y", "x"), np.ones((2, 3))),
            "lat": (("y", "x"), np.ones((2, 3))),
            "time": ("time", [1]),
        },
    )
    ds = cf_specs0.format_dataset(ds0)
    assert "x_u" in ds.dims
    assert "y" in ds.dims
    assert "u" in ds
    assert "bathy_u" in ds


@pytest.mark.parametrize("cf_name", [None, "temp"])
@pytest.mark.parametrize(
    "in_name,in_attrs",
    [
        ("temp", None),
        ("yyy", {"standard_name": "sea_water_temperature"}),
    ],
)
def test_cf_cfspecs_format_data_var(cf_name, in_name, in_attrs):

    lon = xr.DataArray(range(5), dims='xxx', name='xxx', attrs={'standard_name': 'longitude'})
    temp = xr.DataArray(
        range(20, 25), dims='xxx', coords={'xxx': lon}, name=in_name, attrs=in_attrs
    )
    temp = cf.get_cf_specs().format_data_var(temp, cf_name)
    assert temp.name == "temp"
    assert temp.standard_name == "sea_water_temperature"
    assert temp.long_name == "Temperature"
    assert temp.units == "degrees_celsius"
    assert temp.lon.standard_name == "longitude"


def test_cf_cfspecs_format_data_var_coord():
    da = xr.DataArray(0, attrs={'standard_name': 'longitude_at_u_location'})
    da = cf.get_cf_specs().format_data_var(da)


#     assert da.name == "lon_u"


def test_cf_cfspecs_format_data_var_specialize():

    da = xr.DataArray(1, name="salinity")
    cfspecs = cf.CFSpecs({'data_vars': {'sal': {'name': 'supersal'}}})
    da = cfspecs.format_data_var(da, specialize=True)
    assert da.name == "supersal"
    assert da.standard_name == "sea_water_salinity"


def test_cf_cfspecs_format_data_var_loc():
    temp = xr.DataArray(0, name='xtemp', attrs={'standard_name': 'banana_at_x_location'})
    cfspecs = cf.get_cf_specs()

    temp_fmt = cfspecs.format_data_var(temp, "temp", format_coords=False, replace_attrs=True)
    assert temp_fmt.name == "temp"
    assert temp_fmt.standard_name == "sea_water_temperature"  #_at_x_location"

    cfspecs = cf.CFSpecs({"data_vars": {"temp": {"add_loc": True}}})
    temp_fmt = cfspecs.format_data_var(temp, "temp", format_coords=False, replace_attrs=True)
    assert temp_fmt.name == "temp_x"


def test_cf_cfspecs_format_data_var_unkown():
    da = xr.DataArray(range(5), name='foo')
    cfspecs = cf.get_cf_specs()

    da_fmt = cfspecs.format_data_var(da, rename=False)
    assert da_fmt.name == "foo"

    da_fmt = cfspecs.format_data_var(da, rename=True)
    assert da_fmt.name == "foo"


def test_cf_cfspecs_to_loc():
    ds = xr.Dataset(
        {"u_t": (("time", "y", "x_u"), np.ones((1, 2, 3)))},
        coords={
            "lon_u": (("y", "x_u"), np.ones((2, 3))),
            "lat": (("y", "x_u"), np.ones((2, 3))),
            "time": ("time", [1]),
        },
    )
    cfspecs = cf.get_cf_specs()
    dso = cfspecs.to_loc(ds, x=False, y='v', u=None)
    assert "x" in dso.dims
    assert "y_v" in dso.dims
    assert "u_t" in dso
    assert "lon_u" in dso

def test_cf_cfspecs_reloc():
    ds = xr.Dataset(
        {"u_t": (("time", "y", "x_u"), np.ones((1, 2, 3)))},
        coords={
            "lon_u": (("y", "x_u"), np.ones((2, 3))),
            "lat": (("y", "x_u"), np.ones((2, 3))),
            "time": ("time", [1]),
        },
    )
    cfspecs = cf.get_cf_specs()
    dso = cfspecs.reloc(ds, u=False, t='u')
    assert "x" in dso.dims
    assert "u_u" in dso
    assert "lon" in dso

def test_cf_cfspecs_coords_get_axis():
    cfspecs = cf.get_cf_specs().coords

    # from attrs
    depth = xr.DataArray([1], dims='aa', attrs={'axis': 'z'})
    assert cfspecs.get_axis(depth) == 'Z'

    # from CF specs
    depth = xr.DataArray([1], dims='aa', attrs={'standard_name': 'ocean_layer_depth'})
    assert cfspecs.get_axis(depth) == 'Z'


def test_cf_cfspecs_coords_get_dim_type():
    cfspecs = cf.get_cf_specs().coords

    # from name
    assert cfspecs.get_dim_type('aa') is None
    assert cfspecs.get_dim_type('xi') == "x"

    # from a known coordinate
    coord = xr.DataArray([1], dims='aa', attrs={'standard_name': 'longitude'})
    da = xr.DataArray([1], dims='aa', coords={'aa': coord})
    assert cfspecs.get_dim_type('aa', da) == "x"


def test_cf_cfspecs_coords_get_dim_types():
    cfspecs = cf.get_cf_specs().coords

    aa = xr.DataArray([0, 1], dims="aa", attrs={"standard_name": "latitude"})
    da = xr.DataArray(np.ones((2, 2, 2)), dims=('foo', 'aa', 'xi'), coords={'aa': aa})

    assert cfspecs.get_dim_types(da) == (None, 'y', 'x')
    assert cfspecs.get_dim_types(da, unknown='-') == ("-", 'y', 'x')
    assert cfspecs.get_dim_types(da, asdict=True) == {"foo": None, "aa": "y", "xi": "x"}


def test_cf_cfspecs_coords_search_dim():
    cfspecs = cf.get_cf_specs().coords

    # from name
    temp = xr.DataArray(np.arange(2 * 3).reshape(1, 2, 3), dims=('aa', 'ny', 'x'))
    assert cfspecs.search_dim(temp, 'y') == 'ny'
    assert cfspecs.search_dim(temp) is None

    # from explicit axis attribute
    depth = xr.DataArray([1], dims='aa', attrs={'axis': 'z'})
    temp.coords['aa'] = depth
    assert cfspecs.search_dim(temp, 'z') == 'aa'
    assert cfspecs.search_dim(temp) is None
    assert cfspecs.search_dim(depth) == {"dim": "aa", "type": "z", "cf_name": None}

    # from known coordinate
    del temp.coords['aa'].attrs['axis']
    temp.coords['aa'].attrs['standard_name'] = 'ocean_layer_depth'
    assert cfspecs.search_dim(temp, 'z') == 'aa'
    assert cfspecs.search_dim(temp, 'depth') == 'aa'  # by generic name
    assert cfspecs.search_dim(temp) is None

    # subcoords
    level = xr.DataArray(np.arange(2), dims='level')
    depth = xr.DataArray(
        np.arange(2 * 3).reshape(2, 3),
        dims=('level', 'lon'),
        name='depth',
        coords={'level': level, 'lon': [3, 4, 5]},
    )
    assert cfspecs.search_dim(depth) == {'dim': 'level', 'cf_name': 'level', 'type': 'z'}
    assert cfspecs.search_dim(depth.level) == {'dim': 'level', 'type': 'z', 'cf_name': 'level'}
    depth = depth.rename(level='aa')
    depth.aa.attrs['axis'] = 'Z'
    assert cfspecs.search_dim(depth) == {'dim': 'aa', 'type': 'z', 'cf_name': None}

    # not found but only 1d and no dim_type specified
    assert cfspecs.search_dim(xr.DataArray([5], dims='bb')) == {
        'dim': 'bb',
        'type': None,
        'cf_name': None,
    }


def test_cf_cfspecs_coords_search_from_dim():

    lon = xr.DataArray([1, 2], dims='lon')
    level = xr.DataArray([1, 2, 3], dims='aa', attrs={'standard_name': 'ocean_sigma_coordinate'})
    mem = xr.DataArray(range(3), dims='mem')
    temp = xr.DataArray(
        np.zeros((mem.size, level.size, lon.size)),
        dims=('mem', 'aa', 'lon'),
        coords={'mem': mem, 'aa': level, 'lon': lon},
    )

    cfspecs = cf.get_cf_specs().coords

    # Direct coordinate
    assert cfspecs.search_from_dim(temp, 'aa').name == 'aa'

    # Depth coordinate because the only one with this dim
    depth = xr.DataArray(
        np.ones((level.size, lon.size)), dims=('aa', 'lon'), coords={'aa': level, 'lon': lon}
    )
    temp.coords['depth'] = depth
    assert cfspecs.search_from_dim(temp, 'aa').name == 'depth'

    # Cannot get dim_type and we have multiple coords with this dim
    del temp.coords['aa']
    fake = xr.DataArray(np.ones((level.size, lon.size)), dims=('aa', 'lon'), coords={'lon': lon})
    temp.coords['fake'] = fake
    assert cfspecs.search_from_dim(temp, 'aa') is None

    # Can get dim_type back from name
    temp = temp.rename(aa='level')
    assert cfspecs.search_from_dim(temp, 'level').name == 'depth'

    # Nothing identifiable
    temp = xr.DataArray([3], dims='banana')
    assert cfspecs.search_from_dim(temp, "banana") is None


def test_cf_cfspecs_coords_get_dims():
    lat = xr.DataArray([4, 5], dims='yy', attrs={'units': 'degrees_north'})
    depth = xr.DataArray([4, 5], dims='level', attrs={'axis': 'Z'})
    da = xr.DataArray(
        np.ones((2, 2, 2, 2)), dims=('r', 'level', 'yy', 'xi'), coords={'level': depth, 'yy': lat}
    )

    cfspecs = cf.get_cf_specs().coords
    dims = cfspecs.get_dims(da, 'xyzt', allow_positional=True)
    assert dims == ('xi', 'yy', 'level', 'r')
    dims = cfspecs.get_dims(da, 'f', errors="ignore")
    assert dims == (None,)


def test_cf_cfspecs_infer_coords():
    ds = xr.Dataset({"temp": ("nx", [1, 2]), "lon": ("nx", [4, 5])})
    ds = cf.get_cf_specs().infer_coords(ds)
    assert "lon" in ds.coords


def test_cf_cfspecs_decode_encode():
    ds = xoa.open_data_sample("croco.south-africa.meridional.nc")
    cfspecs = cf.CFSpecs(xoa.get_data_sample("croco.cfg"))

    dsc = cfspecs.decode(ds)
    assert list(dsc) == [
        'akt',
        'cs_r',
        'cs_w',
        'Vtransform',
        'angle',
        'el',
        'corio',
        'bathy',
        'hbl',
        'hc',
        'mask_rho',
        'ex',
        'ey',
        'sal',
        'sc_r',
        'sc_w',
        'ptemp',
        'time_step',
        'u',
        'v',
        'w',
        'xl',
        'ssh',
    ]
    assert list(dsc.coords) == [
        'y_rho',
        'y_v',
        'lat_rho',
        'lat_u',
        'lat_v',
        'lon_rho',
        'lon_u',
        'lon_v',
        'sig_rho',
        'sig_w',
        'time',
        'x_rho',
        'x_u',
    ]
    assert set(dsc.dims) == {'auxil', 'sig_rho', 'sig_w', 'time', 'x_rho', 'x_u', 'y_rho', 'y_v'}

    dse = cfspecs.encode(dsc)
    assert list(dse) == list(ds)
    assert list(dse.coords) == list(ds.coords)
    assert list(dse.dims) == list(ds.dims)
    ds.close()


def test_cf_dataarraycfaccessor():
    from xoa.accessors import CFDataArrayAccessor

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", xr.core.extensions.AccessorRegistrationWarning)
        xr.register_dataarray_accessor('xcf')(CFDataArrayAccessor)

    lon = xr.DataArray(range(5), dims='xxx', name='xxx', attrs={'standard_name': 'longitude'})
    temp = xr.DataArray(range(20, 25), dims='xxx', coords={'xxx': lon}, name='temp')

    assert temp.xcf.lon.name == 'xxx'
    assert temp.xcf.lat is None
    assert temp.xcf.lon.xcf.name == "lon"


def test_cf_datasetcfaccessor():
    from xoa.accessors import CFDatasetAccessor

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", xr.core.extensions.AccessorRegistrationWarning)
        xr.register_dataset_accessor('xcf')(CFDatasetAccessor)

    lon = xr.DataArray(range(5), dims='xxx', name='xxx', attrs={'standard_name': 'longitude'})
    temp = xr.DataArray(
        range(20, 25),
        dims='xxx',
        coords={'xxx': lon},
        name='yoyo',
        attrs={'standard_name': 'sea_water_temperature'},
    )

    ds = temp.to_dataset()
    assert ds.xcf.temp.name == 'yoyo'
    assert ds.xcf.sal is None


def test_cf_register_cf_specs():

    cf_cache = cf._get_cache_()
    cf_cache["registered"].clear()

    content = """
        [register]
        name=myname

        [data_vars]
            [[temp]]
            name=mytemp
        """

    cf_specs = cf.CFSpecs(content)
    assert cf_specs.name == "myname"

    cf.register_cf_specs(cf_specs)
    assert cf_specs in cf_cache["registered"]
    assert cf_specs.name == "myname"

    cf.register_cf_specs(myothername=cf_specs)
    assert cf_specs in cf_cache["registered"]
    assert cf_specs.name == "myothername"


def test_cf_get_cf_specs_registered():

    cf_cache = cf._get_cache_()
    cf_cache["registered"].clear()
    content = """
        [register]
        name=myname

        [data_vars]
            [[temp]]
            name=mytemp
        """
    cf_specs_in = cf.CFSpecs(content)
    cf.register_cf_specs(cf_specs_in)

    cf_specs_out = cf.get_cf_specs(name='myname')
    assert cf_specs_out is cf_specs_in


def test_cf_get_cf_specs_from_encoding():

    cf_cache = cf._get_cache_()
    cf_cache["registered"].clear()
    content = """
        [register]
        name=mynam234

        [data_vars]
            [[temp]]
            name=mytemp
        """
    cf_specs_in = cf.CFSpecs(content)
    cf.register_cf_specs(cf_specs_in)

    ds = xr.Dataset(
        {
            "mytemp": (["mylat", "mylon"], np.ones((2, 2))),
            "mysal": (["mylat", "mylon"], np.ones((2, 2))),
        },
        coords={"mylon": np.arange(2), "mylat": np.arange(2)},
    )

    ds.encoding.update(cf_specs="mynam234")
    assert cf.get_cf_specs_from_encoding(ds) is cf_specs_in

    ds.mytemp.encoding.update(cf_specs="mynam234")
    assert cf.get_cf_specs_from_encoding(ds.mytemp) is cf_specs_in

    ds.mylon.encoding.update(cf_specs="mynam234")
    assert cf.get_cf_specs_from_encoding(ds.mylon) is cf_specs_in

    assert cf.get_cf_specs_from_encoding(ds.mylat) is None


def test_cf_set_cf_specs_registered():

    cf_cache = cf._get_cache_()
    cf_cache["registered"].clear()
    content = """
        [register]
        name=myname2

        [data_vars]
            [[temp]]
            name=mytemp
        """
    cf_specs_in = cf.CFSpecs(content)
    cf.register_cf_specs(cf_specs_in)

    with cf.set_cf_specs("myname2") as cfspecs:
        assert cfspecs is cf_specs_in


def test_cf_get_cf_specs_matching_score():

    cf_content0 = """
        [data_vars]
            [[temp]]
            name=mytemp
        """
    cf_specs0 = cf.CFSpecs(cf_content0)
    cf_content1 = """
        [data_vars]
            [[temp]]
            name=mytemp
            [[sal]]
            name=mysal
        [coords]
            [[lon]]
            name=mylon
        """
    cf_specs1 = cf.CFSpecs(cf_content1)
    cf_content2 = """
        [data_vars]
            [[temp]]
            name=mytemp
            [[sal]]
            name=mysal
        """
    cf_specs2 = cf.CFSpecs(cf_content2)

    ds = xr.Dataset(
        {
            "mytemp": (["mylat", "mylon"], np.ones((2, 2))),
            "mysal": (["mylat", "mylon"], np.ones((2, 2))),
        },
        coords={"mylon": np.arange(2), "mylat": np.arange(2)},
    )

    for cf_specs, score in [(cf_specs0, 25), (cf_specs1, 75), (cf_specs2, 50)]:
        assert cf.get_cf_specs_matching_score(ds, cf_specs) == score


def test_cf_infer_cf_specs():

    cf_content0 = """
        [register]
            [[attrs]]
            source="*hycom3d*"

        [data_vars]
            [[temp]]
            name=mytemp
        """
    cf_specs0 = cf.CFSpecs(cf_content0)
    cf_content1 = """
        [data_vars]
            [[temp]]
            name=mytemp
            [[sal]]
            name=mysal
        [coords]
            [[lon]]
            name=mylon
        """
    cf_specs1 = cf.CFSpecs(cf_content1)
    cf_content2 = """
        [register]
        name=hycom3d

        [data_vars]
            [[temp]]
            name=mytemp
            [[sal]]
            name=mysal
        """
    cf_specs2 = cf.CFSpecs(cf_content2)

    cf_cache = cf._get_cache_()
    cf_cache["registered"].clear()
    cf.register_cf_specs(cf_specs0, cf_specs1, cf_specs2)

    temp = xr.DataArray([1], dims="mylon")
    sal = xr.DataArray([1], dims="mylon")
    lon = xr.DataArray([1], dims="mylon")

    ds = xr.Dataset({"mytemp": temp, "mysal": sal}, coords={"mylon": lon})
    assert cf.infer_cf_specs(ds) is cf_specs1

    ds.attrs.update(source="my hycom3d!")
    assert cf.infer_cf_specs(ds) is cf_specs0

    ds.attrs.update(cf_specs="hycom3d")
    assert cf.infer_cf_specs(ds) is cf_specs2


# test_cf_cfspecs_decode_encode()
# test_cf_cfspecs_format_data_var_loc()
# test_cf_cfspecs_coords_get_loc_arg()
# test_cf_cfspecs_format_obj_with_loc()
# test_cf_cfspecs_get_loc_mapping()
# test_cf_cfspecs_coords_search_dim()

# lon = xr.DataArray(range(5), dims="lon")
# banana = xr.DataArray(
#     lon + 20,
#     dims="lon",
#     coords=[lon],
#     name="banana_t",
#     attrs={"standard_name": "banana", "taste": "good"},
# )
# floc, fname, fattrs, out_name, out_standard_name, replace_attrs=None, "sst_q", {"standard_name": ["potatoe"]}, "sst_q", "potatoe_at_q_location", True
# print(cf.SGLocator().format_dataarray(
#     banana, loc=floc, name=fname, attrs=fattrs, replace_attrs=replace_attrs
# ))

# sg = cf.SGLocator()
# print(sg.get_loc(name="u_t", attrs=dict(standard_name = None, long_name = None)))

# test_cf_cfspecs_decode_encode()


"""
Routines related to the ocean dynamics.
"""
# Copyright 2020-2021 Shom
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
import pandas as pd
import xarray as xr

from .__init__ import XoaError
from . import coords as xcoords
from . import grid as xgrid
from xoa.interp import grid2locs
from xoa.geo import EARTH_RADIUS


def _get_uv2d_(t, txy, gx, gy, gz, gt, guv):
    """Interpolate gridded u and v on track positions"""
    # Shape
    npts = int(txy.size / 2)
    tx = txy[:npts]
    ty = txy[npts:]
    tt = np.full(npts, t)
    tz = np.zeros(npts)

    # Interpolate
    tuv = grid2locs(gx, gy, gz, gt, guv, tx, ty, tz, tt)

    # Scale the speed for degrees
    tuv[0] *= 180. / (np.pi * EARTH_RADIUS)
    tuv[1] *= 180. / (np.pi * EARTH_RADIUS)
    tuv[1] *= np.cos(ty*np.pi/180)

    # Pack velocity
    return tuv.ravel()


def _rk4_(xy, f, t, dt, **kwargs):
    """Integrate one time step with RK4"""
    k1 = dt * f(t, xy, **kwargs)
    k2 = dt * f(t + 0.5*dt, xy + 0.5*k1, **kwargs)
    k3 = dt * f(t + 0.5*dt, xy + 0.5*k2, **kwargs)
    k4 = dt * f(t + 0.5*dt, xy + 0.5*k3, **kwargs)
    return xy + (k1 + k2 + k3 + k4)/6.


def _integrate_(xy, f, t0, t1, dt, **kwargs):
    """Low level integration with packed, pure-numeric data"""

    # Fit the time steps to the time range
    dt *= 1.
    if t1 < 0:
        dts = [dt] * int(-t1)
    else:
        nt = int((t1-t0)//dt)
        dts = [dt] * nt
        dtf = (t1-t0) % dt
        if dtf:
            dts.append(dtf)

    # Interative integration
    t = t0
    tt = [t0]
    xxyy = [xy]
    for dt in dts:
        xy2 = _rk4_(xy, f, t, dt, **kwargs)
        t2 = t + dt
        tt.append(t2)
        xxyy.append(xy2)
        t = t2
        xy = xy2
    return np.array(tt), np.asarray(xxyy)


def flow2d(u, v, xy0, duration, step, date=None):
    """Integrate gridded 2D velocities from random positions

    Parameters
    ----------
    u: xarray.DataArray
        Gridded zonal velocity
    v: xarray.DataArray
        Gridded meridional velocity
    duration: int, numpy.timedelta64
        Total integration time in seconds
    step: int, numpy.timedelta64
        Integratiin step in seconds
    xy0: int, xarray.Dataset, tuple
        Either a number of particles or a dataset of initial positions
        with longitude and latitude coordinates
    date: None, numpy.datetime64
        A reference date for the time integration

    Return
    ------
    xarray.Dataset
        Output positions with ``lon`` and ``lat`` coordinates that vary with time.

    """
    # Gridded field
    time0 = xcoords.get_time(u, errors="ignore")
    u = u.squeeze(drop=True)
    v = v.squeeze(drop=True)
    if u.ndim != 2 or v.ndim != 2:
        raise XoaError("The velocity field must 2D")
    u = xgrid.to_rect(u)
    gx = xcoords.get_lon(u).values
    gy = xcoords.get_lat(u).values
    if gx.ndim == 1:
        gx = gx.reshape(1, -1)
    if gy.ndim == 1:
        gy = gy.reshape(-1, 1)
    gz = np.zeros((1,)*5)
    gt = np.zeros(1)
    gu = u.values.reshape((1,)*2+u.shape[-2:])
    gv = v.values.reshape(gu.shape)
    guv = np.array([gu, gv])

    # Initial positions
    tt0 = None
    if isinstance(xy0, xr.Dataset):
        tx0 = xcoords.get_lon(xy0).values.ravel()
        ty0 = xcoords.get_lon(xy0).values.ravel()
        tt0 = xcoords.get_time(xy0, errors='ignore')
    elif isinstance(xy0, int):
        tx0 = np.random.uniform(gx.min(), gx.max(), xy0)
        ty0 = np.random.uniform(gy.min(), gy.max(), xy0)
    else:
        tx0, ty0 = xy0
        tx0 = np.asarray(tx0)
        ty0 = np.asarray(ty0)
    txy0 = np.concatenate((tx0, ty0))

    # Integration
    t0 = 0
    if isinstance(duration, np.timedelta64):
        duration /= np.timedelta64(1, "s")
    if isinstance(step, np.timedelta64):
        step /= np.timedelta64(1, "s")
    t1 = duration
    tt, txy = _integrate_(
        txy0,
        _get_uv2d_,
        t0, t1, step,
        gx=gx, gy=gy, gz=gz, gt=gt,
        guv=guv)
    tx, ty = np.split(txy, 2, axis=1)

    # As a dataset
    if date is not None:
        date = pd.to_datetime(date).to_datetime64()
    elif tt0 is not None and len(tt0) <= 1:
        date = tt0.values
    elif time0 is not None:
        date = time0.values
    else:
        date = pd.Timestamp.now().to_datetime64()
    time = xr.DataArray(tt*np.timedelta64(1, 's')+date, dims='time')

    return xr.Dataset(
        coords={'lon': (('time', 'particles'), tx),
                'lat': (('time', 'particles'), ty),
                'time': time})



# -*- coding: utf-8 -*-
"""
Coordinates and dimensions utilities
"""
# Copyright 2020-2021 Shom
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Mapping

import xarray as xr

from .__init__ import XoaError, xoa_warn
from . import misc
from . import cf as xcf


@misc.ERRORS.format_function_docstring
def get_lon(da, errors="raise"):
    """Get the longitude coordinate

    Parameters
    ----------
    da: xarray.DataArray
    {errors}

    Return
    ------
    xarray.DataArray or None

    See also
    --------
    get_lat
    get_depth
    get_altitude
    get_level
    get_vertical
    get_time
    xoa.cf.CFSpecs.search_coord
    """
    return xcf.get_cf_specs(da).search(da, 'lon', errors=errors)


def is_lon(da, loc="any"):
    """Tell if a data array is identified as longitudes

    Parameters    is_vertical

    da: xarray.DataArray

    Return
    ------
    bool

    See also
    --------
    is_lat
    is_depth
    is_altitude
    is_level
    is_time
    xoa.cf.CFCoordSpecs.match
    """
    return xcf.get_cf_specs(da).coords.match(da, "lon", loc=loc)


@misc.ERRORS.format_function_docstring
def get_lat(da, errors="raise"):
    """Get the latitude coordinate

    Parameters
    ----------
    {errors}

    Return
    ------
    xarray.DataArray or None

    See also
    --------
    get_lon
    get_depth
    get_altitude
    get_level
    get_vertical
    get_time
    xoa.cf.CFSpecs.search_coord
    """
    return xcf.get_cf_specs(da).search(da, 'lat', errors=errors)


def is_lat(da, loc="any"):
    """Tell if a data array is identified as latitudes

    Parameters
    ----------
    da: xarray.DataArray

    Return
    ------
    bool

    See also
    --------
    is_lon
    is_depth
    is_altitude
    is_level
    is_time
    xoa.cf.CFCoordSpecs.match
    """
    return xcf.get_cf_specs(da).coords.match(da, "lat", loc=loc)


@misc.ERRORS.format_function_docstring
def get_depth(da, errors="raise"):
    """Get or compute the depth coordinate

    If a depth variable cannot be found, it tries to compute either
    from sigma-like coordinates or from layer thinknesses.

    Parameters
    ----------
    {errors}

    Return
    ------
    xarray.DataArray or None

    See also
    --------
    get_lon
    get_lat
    get_time
    get_altitude
    get_level
    get_vertical
    xoa.cf.CFSpecs.search_coord
    xoa.sigma.decode_cf_sigma
    xoa.grid.decode_cf_dz2depth
    """
    cfspecs = xcf.get_cf_specs(da)
    errors = misc.ERRORS[errors]
    ztype = cfspecs["vertical"]["type"]

    # From variable
    depth = cfspecs.search(da, 'depth', errors="ignore")
    if depth is not None:
        return depth
    if ztype == "z" or not hasattr(da, "data_vars"):  # explicitly
        msg = "No depth coordinate found"
        if errors == "raise":
            raise XoaError(msg)
        xoa_warn(msg)
        return

    # Decode the dataset
    if ztype == "sigma" or ztype is None:
        err = "ignore" if ztype is None else errors
        from .sigma import decode_cf_sigma
        da = decode_cf_sigma(da, errors=err)
        if "depth" in da:
            return da.depth
    if ztype == "dz2depth" or ztype is None:
        err = "ignore" if ztype is None else errors
        from .grid import decode_cf_dz2depth
        da = decode_cf_dz2depth(da, errors=err)
        if "depth" in da:
            return da.depth
    msg = "Can't infer depth coordinate from dataset"
    if errors == "raise":
        raise XoaError(msg)
    xoa_warn(msg)


def is_depth(da, loc="any"):
    """Tell if a data array is identified as depths

    Parameters
    ----------
    da: xarray.DataArray

    Return
    ------
    bool

    See also
    --------
    is_lon
    is_lat
    is_altitude
    is_level
    is_time
    xoa.cf.CFCoordSpecs.match
    """
    return xcf.get_cf_specs(da).coords.match(da, "depth", loc=loc)


@misc.ERRORS.format_function_docstring
def get_altitude(da, errors="raise"):
    """Get the altitude coordinate

    Parameters
    ----------
    {errors}

    Return
    ------
    xarray.DataArray or None

    See also
    --------
    get_lon
    get_lat
    get_depth
    get_level
    get_vertical
    get_time
    xoa.cf.CFSpecs.search_coord
    """
    return xcf.get_cf_specs(da).search(da, 'altitude', errors=errors)


def is_altitude(da, loc="any"):
    """Tell if a data array is identified as altitudes

    Parameters
    ----------
    da: xarray.DataArray

    Return
    ------
    bool

    See also
    --------
    is_lon
    is_lat
    is_depth
    is_level
    is_time
    xoa.cf.CFCoordSpecs.match
    """
    return xcf.get_cf_specs(da).coords.match(da, "altitude", loc=loc)


@misc.ERRORS.format_function_docstring
def get_level(da, errors="raise"):
    """Get the level coordinate

    Parameters
    ----------
    {errors}

    Return
    ------
    xarray.DataArray or None

    See also
    --------
    get_lon
    get_lat
    get_depth
    get_altitude
    get_time
    xoa.cf.CFSpecs.search_coord
    """
    return xcf.get_cf_specs(da).coords.search(da, 'level', errors=errors)


def is_level(da, loc="any"):
    """Tell if a data array is identified as levels

    Parameters
    ----------
    da: xarray.DataArray

    Return
    ------
    bool

    See also
    --------
    is_lon
    is_lat
    is_depth
    is_altitude
    is_time
    xoa.cf.CFCoordSpecs.match
    """
    return xcf.get_cf_specs(da).coords.match(da, "levels", loc=loc)


@misc.ERRORS.format_function_docstring
def get_vertical(da, errors="raise"):
    """Get either depth or altitude

    Parameters
    ----------
    {errors}

    Return
    ------
    xarray.DataArray or None

    See also
    --------
    get_lon
    get_lat
    get_depth
    get_altitude
    get_level
    get_time
    xoa.cf.CFSpecs.search_coord
    """
    cfspecs = xcf.get_cf_specs()
    height = cfspecs.search(da, 'depth', errors="ignore")
    if height is None:
        height = cfspecs.search(da, 'altitude', errors="ignore")
    if height is None:
        errors = misc.ERRORS[errors]
        msg = "No vertical coordinate found"
        if errors == "raise":
            raise xcf.XoaCFError(msg)
        elif errors == "warn":
            xoa_warn(msg)
    else:
        return height


@misc.ERRORS.format_function_docstring
def get_time(da, errors="raise"):
    """Get the time coordinate

    Parameters
    ----------
    {errors}

    Return
    ------
    xarray.DataArray or None

    See also
    --------
    get_lon
    get_lat
    get_depth
    get_altitude
    get_level
    get_vertical
    xoa.cf.CFSpecs.search_coord
    """
    return xcf.get_cf_specs(da).coords.search(da, 'time', errors=errors)


def is_time(da):
    """Tell if a data array is identified as time

    Parameters
    ----------
    da: xarray.DataArray

    Return
    ------
    bool

    See also
    --------
    is_lon
    is_lat
    is_depth
    is_altitude
    is_level
    xoa.cf.CFCoordSpecs.match
    """
    return xcf.get_cf_specs(da).match(da, "time")


@misc.ERRORS.format_function_docstring
def get_cf_coords(da, coord_names, errors="raise"):
    """Get several coordinates

    Parameters
    ----------
    {errors}

    Return
    ------
    list(xarray.DataArray)

    See also
    --------
    xoa.cf.CFSpecs.search_coord
    """
    cfspecs = xcf.get_cf_specs(da)
    return [cfspecs.search_coord(da, coord_name, errors=errors)
            for coord_name in coord_names]


@misc.ERRORS.format_function_docstring
def get_cf_dims(da, cf_args, allow_positional=False, positions='tzyx', errors="warn"):
    """Get the data array dimensions names from their type

    Parameters
    ----------
    da: xarray.DataArray
        Array to scan
    cf_args: str, list
        Letters among "x", "y", "z", "t" and "f",
        or generic CF names.
    allow_positional: bool
        Fall back to positional dimension of types if unkown.
    positions: str
        Default position per type starting from the end.
    {errors}

    Return
    ------
    tuple
        Tuple of dimension name or None when the dimension if not found

    See also
    --------
    xoa.cf.CFSpecs.get_dims
    """
    return xcf.get_cf_specs(da).get_dims(
        da, cf_args, allow_positional=allow_positional,
        positions=positions, errors=errors)


@misc.ERRORS.format_function_docstring
def get_xdim(da, errors="warn", **kwargs):
    """Get the dimension of X type

    It is a simple call to :func:`get_dims` with ``dim_types="x"``

    Parameters
    ----------
    da: xarray.DataArray
        Array to scan
    positions: str
        Default position per type starting from the end.
    {errors}
    kwargs: dict
        Extra parameters are passed to :func:`get_dims`

    Return
    ------
    str or None
        The dimension name or None

    See also
    --------
    get_dims
    """
    dims = get_cf_dims(da, "x", errors=errors)
    if dims:
        return dims[0]


@misc.ERRORS.format_function_docstring
def get_ydim(da, errors="warn", **kwargs):
    """Get the dimension of Y type

    It is a simple call to :func:`get_dims` with ``dim_types="y"``

    Parameters
    ----------
    da: xarray.DataArray
        Array to scan
    positions: str
        Default position per type starting from the end.
    {errors}
    kwargs: dict
        Extra parameters are passed to :func:`get_dims`

    Return
    ------
    str or None
        The dimension name or None

    See also
    --------
    get_dims
    """
    dims = get_cf_dims(da, "y", errors=errors)
    if dims:
        return dims[0]


@misc.ERRORS.format_function_docstring
def get_zdim(da, errors="warn", **kwargs):
    """Get the dimension of Z type

    It is a simple call to :func:`get_dims` with ``dim_types="z"``

    Parameters
    ----------
    da: xarray.DataArray
        Array to scan
    positions: str
        Default position per type starting from the end.
    {errors}
    kwargs: dict
        Extra parameters are passed to :func:`get_dims`

    Return
    ------
    str or None
        The dimension name or None

    See also
    --------
    get_dims
    """
    dims = get_cf_dims(da, "z", errors=errors)
    if dims:
        return dims[0]


@misc.ERRORS.format_function_docstring
def get_tdim(da, errors="warn", **kwargs):
    """Get the dimension of T type

    It is a simple call to :func:`get_dims` with ``dim_types="t"``

    Parameters
    ----------
    da: xarray.DataArray
        Array to scan
    positions: str
        Default position per type starting from the end.
    {errors}
    kwargs: dict
        Extra parameters are passed to :func:`get_dims`

    Return
    ------
    str or None
        The dimension name or None

    See also
    --------
    get_dims
    """
    dims = get_cf_dims(da, "t", errors=errors)
    if dims:
        return dims[0]


@misc.ERRORS.format_function_docstring
def get_fdim(da, errors="warn", **kwargs):
    """Get the dimension of F type

    It is a simple call to :func:`get_dims` with ``dim_types="f"``

    Parameters
    ----------
    da: xarray.DataArray
        Array to scan
    positions: str
        Default position per type starting from the end.
    {errors}
    kwargs: dict
        Extra parameters are passed to :func:`get_dims`

    Return
    ------
    str or None
        The dimension name or None

    See also
    --------
    get_dims
    """
    dims = get_cf_dims(da, "f", errors=errors)
    if dims:
        return dims[0]


class transpose_modes(misc.IntEnumChoices, metaclass=misc.DefaultEnumMeta):
    """Supported :func:`transpose` modes"""
    #: Basic xarray transpose with :meth:`xarray.DataArray.transpose`
    classic = 0
    basic = 0
    xarray = 0
    #: Transpose skipping incompatible dimensions
    compat = -1
    #: Transpose adding missing dimensions with a size of 1
    insert = 1
    #: Transpose resizing to missing dimensions.
    #: Note that dims must be an array or a dict of sizes
    #: otherwise new dimensions will have a size of 1.
    resize = 2


def transpose(da, dims, mode='compat'):
    """Transpose an array

    Parameters
    ----------
    da: xarray.DataArray
        Array to tranpose
    dims: tuple(str), xarray.DataArray, dict
        Target dimensions or array with dimensions
    mode: str, int
        Transpose mode as one of the following:
        {transpose_modes.rst_with_links}

    Return
    ------
    xarray.DataArray
        Transposed array

    Example
    -------
    .. ipython:: python

        @suppress
        import xarray as xr, numpy as np
        @suppress
        from xoa.coords import transpose
        a = xr.DataArray(np.ones((2, 3, 4)), dims=('y', 'x', 't'))
        b = xr.DataArray(np.ones((10, 3, 2)), dims=('m', 'y', 'x'))
        # classic
        transpose(a, (Ellipsis, 'y', 'x'), mode='classic')
        # insert
        transpose(a, ('m', 'y', 'x', 'z'), mode='insert')
        transpose(a, b, mode='insert')
        # resize
        transpose(a, b, mode='resize')
        transpose(a, b.sizes, mode='resize') # with dict
        # compat mode
        transpose(a, ('y', 'x'), mode='compat').dims
        transpose(a, b.dims, mode='compat').dims
        transpose(a, b, mode='compat').dims  # same as with b.dims

    See also
    --------
    xarray.DataArray.transpose
    """
    # Inits
    if hasattr(dims, 'dims'):
        sizes = dims.sizes
        dims = dims.dims
    elif isinstance(dims, Mapping):
        sizes = dims
        dims = list(dims.keys())
    else:
        sizes = None
    mode = str(transpose_modes[mode])
    kw = dict(transpose_coords=True)

    # Classic
    if mode == "classic":
        return da.transpose(*dims, **kw)

    # Get specs
    odims = ()
    expand_dims = {}
    with_ell = False
    for dim in dims:
        if dim is Ellipsis:
            with_ell = True
            odims += dim,
        elif dim in da.dims:
            odims += dim,
        elif mode == "insert":
            expand_dims[dim] = 1
            odims += dim,
        elif mode == "resize":
            if sizes is None or dim not in sizes:
                xoa_warn(f"new dim '{dim}' in transposition is set to one"
                         " since no size is provided to it")
                size = 1
            else:
                size = sizes[dim]
            expand_dims[dim] = size
            odims += dim,

    # Expand
    if expand_dims:
        da = da.expand_dims(expand_dims)

    # Input dimensions that were not specified in transposition
    # are flushed to the left
    if not with_ell and set(odims) < set(da.dims):
        odims = (...,) + odims

    # Transpose
    return da.transpose(*odims, **kw)


transpose.__doc__ = transpose.__doc__.format(**locals())


def get_dim_types(da, unknown=None, asdict=False):
    """Get dimension types

    Parameters
    ----------
    da: xarray.DataArray or tuple(str)
        Data array or tuple of dimensions
    unknown:
        Value to assign to unknown types
    asdict: bool
        Get the result as dictionary

    Return
    ------
    tuple
    """
    return xcf.get_cf_specs(da).coords.get_dim_types(
        da, unknown=unknown, asdict=asdict)


def get_order(da):
    """Like :func:`get_dim_types` but returning a string"""
    return "".join(get_dim_types(da, unknown="-", asdict=False))


def reorder(da, order):
    """Transpose an array to match a given order

    Parameters
    ----------
    da: xarray.DataArray
        Data array to transpose
    order: str
        A combination of x, y, z, t, f and - symbols and
        their upper case value.
        Letters refer to the dimension type.
        When the value is -, it may match any dimension type.

    Return
    ------
    xarray.DataArray
    """
    # Convert from dim_types
    if isinstance(order, dict):
        order = tuple(order.values())
    if isinstance(order, tuple):
        order = ''.join([
            ('-' if o not in "ftzyx" else o) for o in order])

    # From order to dims
    to_dims = ()
    dim_types = get_dim_types(da, asdict=True)
    ndim = len(dim_types)
    for i, o in enumerate(order[::-1]):
        if i+1 == ndim:
            break
        for dim in da.dims:
            if o == dim_types[dim]:
                to_dims = (dim, ) + to_dims
                break
        else:
            raise XoaError(
                f"Coordinate type not found: {o}. Dims are: {da.dims}")

    # Final transpose
    return transpose(da, to_dims)


def get_coords_compat_with_dims(da, include_dims=None, exclude_dims=None):
    """Return the coordinates that are compatible with dims

    Parameters
    ----------
    da: xarray.DataArray
        Data array
    include_dims: set(str)
        If provided, the coordinates must have at least one of these
        dimensions
    exclude_dims: set(str)
        If provided, the coordinates must not have one of these dimnesions

    Return
    ------
    list(str)
        List of coordinates
    """
    if isinstance(include_dims, str):
        include_dims = {include_dims}
    if isinstance(exclude_dims, str):
        exclude_dims = {exclude_dims}
    coords = []
    for coord in da.coords.values():
        dims = set(coord.dims)
        if include_dims and not include_dims.intersection(dims):
            continue
        if exclude_dims and exclude_dims.intersection(dims):
            continue
        coords.append(coord)
    return coords


def change_index(da, dim, values):
    """Change the values of a dataset or data array index

    Parameters
    ----------
    da: xarray.Dataset, xarray.DataArray
    dim: str
    values: array_like

    Return
    ------
    xarray.Dataset, xarray.DataArray

    See also
    --------
    xarray.DataArray.reset_index
    xarray.DataArray.assign_coords
    """
    attrs = da.coords[dim].attrs
    if hasattr(values, "attrs"):
        attrs.update(attrs)
    if dim in da.indexes:
        da = da.reset_index([dim], drop=True)
    coord = xr.DataArray(values, dims=dim, attrs=attrs)
    return da.assign_coords({dim: coord})


def drop_dim_coords(da, dim):
    """Drop coords that have a particular dim"""
    return da.drop([c.name for c in da.coords.values() if dim in c.dims])


class positive_attr(misc.IntEnumChoices, metaclass=misc.DefaultEnumMeta):
    """Allowed value for the positive attribute argument"""
    #: Infer it from the axis coordinate
    infer = 0
    guess = 0
    #: Coordinates are increasing up
    up = 1
    #: Coordinates are increasing down
    down = -1


def get_positive_attr(da, zdim=None):
    """Get the positive attribute of a dataset

    Parameters
    ----------
    da: xarray.Dataset, xarray.DataArray
    zdim: None, str
        The index coordinate name that is supposed to have this attribute,
        which is usually the vertical dimension

    Return
    ------
    None, "up" or "down"
    """
    # Targets
    if zdim is None:
        zdim = get_cf_dims(da, "z", errors="ignore")
        if zdim:
            zdim = zdim[0]
    if zdim and zdim in da.coords:
        targets = [da.coords[zdim]]
    else:
        targets = list(da.coords.values())
        if isinstance(da, xr.Dataset):
            targets.extend(da.data_vars.values())

    # Loop on targets
    for target in targets:
        if "positive" in target.attrs:
            positive = da.coords[zdim].attrs["positive"]
            return positive_attr[positive].name

    # Fall back to current CFSpecs
    cfspecs = xcf.get_cf_specs(da)
    return cfspecs["vertical"]["positive"]


def get_binding_data_vars(ds, coord, as_names=False):
    """Get the data_vars that have this coordinate

    Parameters
    ----------
    ds: xarray.Dataset
    coord_name: str

    Return
    ------
    list
        List of data_var names
    """
    if not isinstance(coord, str):
        coord = coord.name
    out = [da for da in ds if coord in da.coords]
    if as_names:
        out = [da.name for da in out]
    return out



"""Generate files to declare and list default CF specifications"""

import os
import logging

import xoa.cf


skip_keys = ["inherit", "processed"]

cat_titles = {
    "data_vars": "Data variables (:attr:`~xoa.cf.CFSpecs.data_vars`)",
    "coords": "Coordinates (:attr:`~xoa.cf.CFSpecs.coords`)",
    "dims": "Dimensions (:attr:`~xoa.cf.CFSpecs.dims`)",
    'sections': 'Other sections',
    }

roles = {
    "data_vars": "cfdatavar",
    "coords": "cfcoord"
    }


def add_items(rst, data, indent=0, subformat='[`{key}`]'):
    local = ""
    nest = {}
    for key, value in data.items():
        if key in skip_keys:
            continue
        if isinstance(value, dict):
            nest[key] = value
        else:
            if isinstance(value, list):
                value = " ".join(f"``{v!r}``" for v in value)
            else:
                value = f"``{value!r}``"
            local += "\t"*indent + f"    * - **{key}**\n"
            local += "\t"*indent + f"      - {value}\n"
    if local:
        rst += "\t"*indent + ".. list-table::\n\n" + local

    for key, dvalue in nest.items():
        rst += "\n" + "\t"*indent + f'.. rubric:: [``"{key}"``]\n\n'
        if dvalue:
            rst = add_items(rst, dvalue, indent=indent)
        else:
            rst += "\t"*indent + "Empty dictionary\n\n"
    return rst


def genrst(app):

    srcdir = app.env.srcdir

    gendir = os.path.join(srcdir, "gencfspecs")
    if not os.path.exists(gendir):
        os.makedirs(gendir)

    cfspecs = xoa.cf.get_cf_specs()
    comments = xoa.cf._get_cfgm_().specs.inline_comments

    rst_tables = {}
    rst_toctrees = {}

    # Data vars and coordinates
    for cfcat in ["data_vars", "coords"]:

        rst_tables[cfcat] = ".. list-table::\n\n"
        rst_toctrees[cfcat] = ".. toctree::\n    :hidden:\n\n"

        logging.info(f"Generating rst files for xoa.cf {cfcat} specs")
        for cfname in cfspecs[cfcat]:

            decdir = os.path.join(gendir, cfcat)
            if not os.path.exists(decdir):
                os.makedirs(decdir)

            # Single declaration file
            role = roles[cfcat]
            with open(os.path.join(decdir, cfname+".rst"), "w") as f:
                title = f':attr:`~xoa.cf.CFSpecs.{cfcat}` [``"{cfname}"``]'
                title += "\n" + len(title) * "="
                rst = title + "\n\n"
                rst += f".. {role}:: {cfname}\n\n"
                rst = add_items(rst, cfspecs[cfcat][cfname], indent=1)
                f.write(rst)

            # Append to table
            rst_tables[cfcat] += f"    * - :{role}:`{cfname}`\n"
            rst_tables[cfcat] += "      - {}\n".format(
                cfspecs[cfcat][cfname]["attrs"]["long_name"][0])

            # Append to toctree
            rst_toctrees[cfcat] += f"    gencfspecs/{cfcat}/{cfname}\n"

    # Dimensions
    rst_tables["dims"] = ".. list-table::\n\n"
    for dim_type, dims in cfspecs["dims"].items():
        rst_tables["dims"] += f"    * - :cfdim:`{dim_type}`\n"
        rst_tables["dims"] += "      - {}\n".format(
            ", ".join([f"``{dim}``" for dim in dims]))

    # Other sections
    rst_tables["sections"] = ".. list-table::\n\n"
    rst_toctrees["sections"] = ".. toctree::\n    :hidden:\n\n"
    for section in "register", "sglocator", "vertical", "accessors":

        logging.info(f"Generating rst files for xoa.cf {section} specs")
        with open(os.path.join(gendir, section+".rst"), "w") as f:
            title = f':attr:`~xoa.cf.CFSpecs` [``"{section}"``]'
            title += "\n" + len(title) * "="
            rst = title + "\n\n"
            rst += f".. cfsec:: {section}\n\n"
            rst = add_items(rst, cfspecs[section], indent=1)
            f.write(rst)

        # Append to table
        rst_tables["sections"] += f"    * - :cfsec:`{section}`\n"
        rst_tables["sections"] += "      - {}\n".format(
            comments[section].strip("# "))

        # Append to toctree
        rst_toctrees["sections"] += f"    gencfspecs/{section}\n"

    # Write the index.txt
    with open(os.path.join(gendir, "index.txt"), "w") as f:

        for key, title in cat_titles.items():

            # Title
            title = title + "\n" + len(title)*"^"
            f.write(f".. _appendix.cf.{key}:\n\n" + title + "\n\n")

            # Hidden toctree
            if key in rst_toctrees:
                f.write(rst_toctrees[key]+"\n")

            # Table
            f.write(rst_tables[key]+"\n\n")



def setup(app):

    app.add_object_type(
        'cfdatavar', 'cfdatavar',
        objname='xoa.cf CFSpecs.data_vars item',
        indextemplate='pair: %s; xoa.cf CFSpecs.data_vars item')

    app.add_object_type(
        'cfcoord', 'cfcoord',
        objname='xoa.cf CFSpecs.coords item',
        indextemplate='pair: %s; xoa.cf CFSpecs.coords item')

    app.add_object_type(
        'cfdim', 'cfdim',
        objname='xoa.cf CFSpecs.coords.dims item',
        indextemplate='pair: %s; xoa.cf CFSpecs.coords.dims item')

    app.add_object_type(
        'cfsec', 'cfsec',
        objname='xoa.cf CFSpecs section',
        indextemplate='pair: %s; xoa.cf CFSpec section')

    app.connect('builder-inited', genrst)

    return {'version': '0.1'}



# -*- coding: utf-8 -*-
"""
Terrain following parametric vertical coordinates

This follows the CF conventions for
`Parametric Vertical Coordinates <http://cfconventions.org/Data/cf-conventions/cf-conventions-1.8/cf-conventions.html#parametric-v-coord>`_.

"""
# Copyright 2020-2021 Shom
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import numpy as np

from .__init__ import xoa_warn
from . import misc
from . import cf
from . import coords as xcoords

#: To convert from formula terms to CF names
FORMULA_TERMS_TO_CF_NAMES = {
    'c': 'cs',  # C
    's': 'sig',
    'sigma': 'sig',
    'eta': 'ssh',
    'depth': 'bathy',
    'depth_c': 'hc',
    'a': 'thetas',
    'b': 'thetab',
}

#: CF names that are known to have horizontal dimensions
HORIZONTAL_TERMS = ["ssh", "bathy", "hc"]

#: Supported sigma coordinates
SIGMA_COORDINATE_TYPES = (
    "atmosphere_sigma_coordinate",
    "ocean_sigma_coordinate",
    "ocean_s_coordinate",
    "ocean_s_coordinate_g1",
    "ocean_s_coordinate_g2",
)


class XoaSigmaError(cf.XoaCFError):
    pass


def atmosphere_sigma_coordinate(sig, ps, ptop, cache=None):
    """Convert from sigma [0, 1] to altitude in an atmopsheric model

    Source:
        `Ocean sigma coordinate  <http://cfconventions.org/Data/cf-conventions/cf-conventions-1.8/cf-conventions.html#_atmosphere_sigma_coordinate>`_

    Formula:
        .. math::

            p = p_{top} + \\sigma * (p_{surf}-p_{top})

    Sigma standard name:
        ``atmosphere_sigma_coordinate``

    Formula terms:
        ``sigma: var1 ps: var2 ptop: var3``

    Parameters
    ----------
    sig: xarray.DataArray
        Sigma coordinates range from 0 to 1 (:math:`\\sigma` | ``sigma``)
    ps: xarray.DataArray
        Surface air pressure (:math:`p_{surf}` | ``ps``)
    ptop: xarray.DataArray
        Air pressure at top of model (:math:`p_{top}` | ``ptop``)

    Returns
    -------
    xarray.DataArray
        Air pressure in Pa (:math:`p`)
    """
    # Compute
    p = sig * (ps - ptop)
    p = p + ptop

    # Format
    return cf.get_cf_specs(sig).format_data_var(p, "plev", format_coords=False)


def ocean_sigma_coordinate(sig, ssh, bathy, cache=None):
    """Convert from sigma [-1, 0] to negative depths in an ocean model

    Source:
        `Ocean sigma coordinate <http://cfconventions.org/Data/cf-conventions/cf-conventions-1.8/cf-conventions.html#_ocean_sigma_coordinate>`_

    Formula:
        .. math::

            z = \\eta + \\sigma * (\\eta+h)

    Sigma standard name:
        ``ocean_sigma_coordinate``

    Formula terms:
        ``sigma: var1 eta: var2 depth: var3``

    Parameters
    ----------
    sig: xarray.DataArray
        Sigma coordinates range from 0 to 1 (:math:`\\sigma` | ``sigma``)
    ssh: xarray.DataArray
        Surface air pressure (:math:`\\eta` | ``eta``)
    bathy: xarray.DataArray
        Positive sea floor depth (:math:`h` | ``depth``)

    Returns
    -------
    xarray.DataArray
        Negative depth below surface in m (:math:`z`)
    """
    # Compute
    if not set(bathy.dims) <= set(ssh.dims):
        raise XoaSigmaError("Incompatible dimensions between bathy and ssh")
    z = sig * (bathy + ssh)
    z = z + ssh

    # Format
    return cf.get_cf_specs(sig).format_data_var(z, "depth", format_coords=False)


def get_cs(sig, thetas, thetab, cs_type=None):
    """Get a s-coordinate stretching curve

    Parameters
    ----------
    sig: xarray.DataArray
        Sigma coordinates range from 0 to 1 (:math:`s` | ``s``)
    thetas: xarray.DataArray
        Surface control parameter (:math:`a` | ``a``)
    thetab: xarray.DataArray
        Bottom control parameter (:math:`b` | ``b``)
    cs_type: str, None
        Stretching type:
            ``None`` (default):

             .. math::

                 C & = (1-b)*\\frac{\\sinh(a*s)}{\\sinh(a)} +  b*\\left[\\frac{\\tanh(a*(s+0.5))}{2*\\tanh(0.5*a)} - 0.5\\right]

    Returns
    -------
    xarray.DataArray
        Stretching curve (:math:`C` | ``C``)
    """
    s, a, b = sig, thetas, thetab
    cs = np.sinh(s * a) * (1 - b) / np.sinh(a)
    cs = cs + b * (np.tanh(a * (s + 0.5)) / (2 * np.tanh(0.5 * a)) - 0.5)
    return cs


def ocean_s_coordinate(sig, ssh, bathy, hc, thetas, thetab, cs=None, cs_type=None, cache=None):
    """Convert from s [-1, 0] to depths in an ocean model

    Source:
        `Ocean s-coordinate <http://cfconventions.org/Data/cf-conventions/cf-conventions-1.8/cf-conventions.html#_ocean_s_coordinate>`_

    Formula:
        .. math::

            z & = \\eta*(1+s) + h_c*s + (h-h_c)*C

            C & = (1-b)*\\frac{\\sinh(a*s)}{\\sinh(a)} +  b*\\left[\\frac{\\tanh(a*(s+0.5))}{2*\\tanh(0.5*a)} - 0.5\\right]

    Sigma standard name:
        ``ocean_s_coordinate``

    Formula terms:
        ``s: var1 eta: var2 depth: var3 a: var4 b: var5 depth_c: var6``

    Parameters
    ----------
    sig: xarray.DataArray
        Sigma coordinates range from -1 to 0 (:math:`s` | ``s``)
    ssh: xarray.DataArray
        Surface air pressure (:math:`\\eta` | ``eta``)
    bathy: xarray.DataArray
        Positive sea floor depth (:math:`h` | ``depth``)
    hc: xarray.DataArray, float
        Positive critical depth (:math:`h_c` | ``depth_c``)
    thetas: xarray.DataArray
        Surface control parameter (:math:`a` | ``a``)
    thetab: xarray.DataArray
        Bottom control parameter (:math:`b` | ``b``)
    cs: xarray.DataArray, None
        Stretching curve, which defaults to the formula above
        computed by :func:`get_cs` (:math:`C` | ``C``)
    cs_type: str, None
        Stretching type (see :func:`get_cs`)
    cache: dict
        Dict variable that stores intermediate results to be used
        from call to call.

    Returns
    -------
    xarray.DataArray
        Negative depth below surface in m (:math:`z`)
    """
    # Check cache first
    if cache:
        zconst = cache['zconst']

    else:  # Compute

        # Stetching curve
        if cs is None:
            cs = get_cs(sig, thetas, thetab, cs_type)

        # Constant part of the formula
        if np.ndim(hc) and set(hc.dims) != set(bathy.dims):
            raise XoaSigmaError("Incompatible dimensions between bathy and hc")
        zconst = sig * hc
        zconst = zconst + cs * (bathy - hc)

        # Store in cache
        if isinstance(cache, dict):
            cache.update(zconst=zconst)

    # Final computation
    z = (sig + 1) * ssh + zconst

    # Format
    return cf.get_cf_specs(sig).format_data_var(z, "depth", format_coords=False)


def ocean_s_coordinate_g1(
    sig, ssh, bathy, hc, thetas=None, thetab=None, cs=None, cs_type=None, cache=None
):
    """Convert from s [-1, 0] generic form 1 to depths in an ocean model

    Source:
        `Ocean s-coordinate, generic form 1 <http://cfconventions.org/cf-conventions/cf-conventions.html#_ocean_s_coordinate_generic_form_1>`_

    Formula:
        .. math::

            z & = \\eta*(1+s) + (\\eta + h * S)

            S & = \\frac{h_c s + h C}{h_c + h}

            C & = (1-b)*\\frac{\\sinh(a*s)}{\\sinh(a)} +  b*\\left[\\frac{\\tanh(a*(s+0.5))}{2*\\tanh(0.5*a)} - 0.5\\right]



    Sigma standard name:
        ``ocean_s_coordinate_g2``

    Formula terms:
        ``s: var1 C: var2 eta: var3 depth: var4 depth_c: var5``

    Parameters
    ----------
    sig: xarray.DataArray
        Sigma coordinates range from -1 to 0 (:math:`s` | ``s``)
    ssh: xarray.DataArray
        Surface air pressure (:math:`\\eta` | ``eta``)
    bathy: xarray.DataArray
        Positive sea floor depth (:math:`h` | ``depth``)
    hc: xarray.DataArray, float
        Positive critical depth (:math:`h_c` | ``depth_c``)
    thetas: xarray.DataArray
        Surface control parameter (:math:`a` | ``a``)
        Optional if `cs` is provided.
    thetab: xarray.DataArray
        Bottom control parameter (:math:`b` | ``b``)
        Optional if `cs` is provided.
    cs: xarray.DataArray, None
        Stretching curve, which defaults to the formula above
        computed by :func:`get_cs` (:math:`C` | ``C``)
    cs_type: str, None
        Stretching type (see :func:`get_cs`)
    cache: dict
        Dict variable that stores intermediate results to be used
        from call to call.

    Returns
    -------
    xarray.DataArray
        Negative depth below surface in m (:math:`z`)
    """
    s, h = sig, bathy

    # Check cache first
    if cache:
        S = cache['S']
        Sh1 = cache['Sh1']

    else:  # Compute

        # Stetching curve
        if cs is None:
            if None in (thetas, thetab):
                raise XoaSigmaError("thetas and thetas must be provided when " "cs is not")
            cs = get_cs(sig, thetas, thetab, cs_type)

        # Constant part of the formula
        if np.ndim(hc) and set(hc.dims) != set(h.dims):
            raise XoaSigmaError("Incompatible dimensions between bathy and hc")
        S = s * hc + cs * (h - hc)
        Sh1 = S / h
        Sh1 = Sh1 + 1

        # Store in cache
        if isinstance(cache, dict):
            cache.update(S=S, Sh1=Sh1)

    # Final computation
    z = S + Sh1 * ssh

    # Format
    return cf.get_cf_specs(sig).format_data_var(z, "depth", format_coords=False)


def ocean_s_coordinate_g2(
    sig, ssh, bathy, hc, thetas=None, thetab=None, cs=None, cs_type=None, cache=None
):
    """Convert from s [-1, 0] generic form 2 to depths in an ocean model

    Source:
        `Ocean s-coordinate, generic form 2 <http://cfconventions.org/Data/cf-conventions/cf-conventions-1.8/cf-conventions.html#_ocean_s_coordinate_generic_form_2>`_

    Formula:
        .. math::

            z & = \\eta*(1+s) + (\\eta + h * S)

            S & = \\frac{h_c s + h C}{h_c + h}

            C & = (1-b)*\\frac{\\sinh(a*s)}{\\sinh(a)} +  b*\\left[\\frac{\\tanh(a*(s+0.5))}{2*\\tanh(0.5*a)} - 0.5\\right]

    Sigma standard name:
        ``ocean_s_coordinate_g2``

    Formula terms:
        ``s: var1 C: var2 eta: var3 depth: var4 depth_c: var5``

    Parameters
    ----------
    sig: xarray.DataArray
        Sigma coordinates range from -1 to 0 (:math:`s` | ``s``)
    ssh: xarray.DataArray
        Surface air pressure (:math:`\\eta` | ``eta``)
    bathy: xarray.DataArray
        Positive sea floor depth (:math:`h` | ``depth``)
    hc: xarray.DataArray, float
        Positive critical depth (:math:`h_c` | ``depth_c``)
    thetas: xarray.DataArray
        Surface control parameter (:math:`a` | ``a``).
        Optional if `cs` is provided.
    thetab: xarray.DataArray
        Bottom control parameter (:math:`b` | ``b``)
        Optional if `cs` is provided.
    cs: xarray.DataArray, None
        Stretching curve, which defaults to the formula above
        computed by :func:`get_cs` (:math:`C` | ``C``)
    cs_type: str, None
        Stretching type (see :func:`get_cs`)
    cache: dict
        Dict variable that stores intermediate results to be used
        from call to call.

    Returns
    -------
    xarray.DataArray
        Negative depth below surface in m (:math:`z`)
    """
    s, h = sig, bathy

    # Check cache first
    if cache:
        S = cache['S']
        hS = cache['hS']

    else:  # Compute

        # Stetching curve
        if cs is None:
            if None in (thetas, thetab):
                raise XoaSigmaError("thetas and thetas must be provided when cs is not")
            cs = get_cs(sig, thetas, thetab, cs_type)

        # Constant part of the formula
        if np.ndim(hc) and set(hc.dims) != set(h.dims):
            raise XoaSigmaError("Incompatible dimensions between bathy and hc")
        S = s * hc + cs * h
        S = S / (hc + h)
        hS = S * h

        # Store in cache
        if isinstance(cache, dict):
            cache.update(S=S, hS=hS)

    # Final computation
    z = hS + (S + 1) * ssh

    # Format
    return cf.get_cf_specs(sig).format_data_var(z, "depth", format_coords=False)


def _ds_search_ci_(ds, name):
    """Case insensitive search in data_vars and coords of a dataset

    Parameters
    ----------
    ds: xarray.Dataset
    name: str
        Requested name

    Returns
    -------
    str
        Real name
    """
    lname = name.lower()
    for cat in "data_vars", "coords":
        pool = getattr(ds, cat)
        names = [nm for nm in pool.keys()]
        lnames = [nm.lower() for nm in names]
        if lname in lnames:
            return names[lnames.index(lname)]


_re_ft_split_terms = re.compile(r'\b\s+\b').split

_re_ft_split_item = re.compile(r'\s*:\s*').split


def decode_formula_terms(attr):
    """Parse the formula_terms attribute

    Parameters
    ----------
    attr: str
        Attribute value

    Returns
    -------
    dict

    Example
    -------
    .. ipython:: python

        @suppress
        from xoa.sigma import decode_formula_terms
        decode_formula_terms(
            's: sc_r C: Cs_r eta: zeta depth: h depth_c: hc')
    """
    terms = {}
    for item in _re_ft_split_terms(attr):
        item = _re_ft_split_item(item)
        if len(item) != 2:
            raise XoaSigmaError("Malformed formula_terms attribute: " + attr)
        terms[item[0]] = item[1]
    return terms


def get_sigma_terms(ds, vloc=None, hlocs=None, rename=False):
    """Get sigma terms from a dataset as another dataset

    It operates like this:

    1. Search for the sigma variables.
    2. Parse their ``formula_terms`` attribute.
    3. Create a dict for each locations from names in datasets to
       :mod:`xoa.cf` compliant names that are also used in conversion
       functions.

    Parameters
    ----------
    ds: xarray.Dataset
    vloc: str, {"any", None}, False
        Staggered grid vertical location.
        If any or None, results for all locations are returned.
    hlocs: None, list of str
        A list of horizontal grid locations

    Returns
    -------
    dict, dict of dict, dict of dict of dict
        A dict is generated for a given sigma variable,
        whose keys are array names, like ``"sc_r"``,
        and values are :mod:`~xoa.cf` names, like ``"sig"``.
        A special key is the ``"type"`` whose corresponding value
        is the ``standard_name``, stripped from its potential staggered grid
        location indicator.
        If ``loc`` is ``"any"`` or ``None``,
        each dict is embedded in a master dict
        whose keys are staggered grid location. If no location is found,
        the key is set ``None``.

    Raises
    ------
    xoa.sigma.XoaSigmaError
        In case of:

        - inconsistent staggered grid location in dataarrays
          as checked by :meth:`xoa.cf.SGLocator.get_location`
        - no standard_name in sigma/s variable
        - a malformed formula
        - a formula term variable that is not found in the dataset
        - an unknown formula term name
    """
    # Get sigma arrays
    cfspecs = cf.get_cf_specs(ds)
    vsingle = vloc not in ("any", None)
    hsingle = not isinstance(hlocs, list)
    if hsingle:
        hlocs = [hlocs]
    sigs = cfspecs.search(ds, 'sig', loc=vloc, single=False)
    terms = {}
    for sig in sigs:

        # Check standard_name and get loc
        if "standard_name" not in sig.attrs:
            raise XoaSigmaError(
                "No standard_name attribute found in sigma/s " "variable name: " + sig.name
            )
        standard_name, loc = cfspecs.sglocator.parse_attr("standard_name", sig.standard_name)
        if standard_name not in SIGMA_COORDINATE_TYPES:
            raise XoaSigmaError(
                "Sigma/s coordinate not supported: "
                + standard_name
                + ". Supported coordinates: "
                + " ".join(SIGMA_COORDINATE_TYPES)
            )

        # Get formula terms
        if "formula_terms" not in sig.attrs:
            raise XoaSigmaError(
                f"Sigma/s type variable {sig.name} " "has no formula_term attribute"
            )
        formula_terms = decode_formula_terms(sig.formula_terms)

        # Loop in horizontal locations
        subterms = terms[loc] = {}
        for hloc in hlocs:

            # Check terms
            subsubterms = subterms[hloc] = {sig.name: "sig", "type": standard_name}
            for fname, fvname in formula_terms.items():

                # xoa.cf name
                # TODO: handle mising cs and fallback with thetas and thetab
                if fname.lower() not in FORMULA_TERMS_TO_CF_NAMES:
                    raise XoaSigmaError("Unknown formula term name: " + fname)
                cf_name = FORMULA_TERMS_TO_CF_NAMES[fname.lower()]

                # Real name
                if cf_name in HORIZONTAL_TERMS:
                    fvname = cfspecs.sglocator.format_attr("name", fvname, hloc)
                vname = _ds_search_ci_(ds, fvname)
                if vname is None:
                    raise XoaSigmaError("Formula array not found: " + fvname)

                subsubterms[vname] = cf_name

    if hsingle:
        for subterms in terms.values():
            subsubterms = subterms[hloc]
            del subterms[hloc]
            subterms.update(subsubterms)
    if vsingle:
        return subterms if sigs else None
    return terms

    # # Rename terms in dict or ds
    # ds = ds.rename({sigma.name: "sig"})
    # for term_name, da_name_ in terms.items():
    #     da_name = _ds_search_ci_(ds, da_name_)
    #     if da_name is None:
    #         xoa_warn("Formula term dataarray not found: " + da_name_)
    #     ds = ds.rename({da_name: term_name})
    # return ds


@misc.ERRORS.format_function_docstring
def decode_cf_sigma(ds, rename=False, hlocs=None, errors="raise"):
    """Compute heights from sigma-like variable in a dataset

    This makes use of the :meth:`~xoa.cf.CFSpecs` instance that is retreived
    with :func:`xoa.cf.get_cf_specs` with ds as an argument in order to
    find needed variables.
    If the dataset is not found to have sigma-like coordinates,
    a simple copy is returned.

    When a data variable that have the same dimensions is found, the
    the computed coordinate is transposed properly and assigned
    to the variable as a a coordinate array.

    Parameters
    ----------
    ds: xarray.Dataset
        Dataset that contains everything needed to compute heights
    rename: bool
        Rename and format arrays ot make them compliant with
        :mod:`xoa.cf`
    hlocs: None, list of str
        Horizontal staggered grid locations to search for "bathy" and "ssh".
        By default, no location is assumed.
    {errors}

    Return
    ------
    xarray.Dataset
    """
    # Init
    ds = ds.copy()
    cfspecs = cf.get_cf_specs(ds)
    errors = misc.ERRORS[errors]

    # Decode formula terms
    try:
        all_terms = get_sigma_terms(ds, vloc=None, hlocs=hlocs)
        if all_terms is None:
            return ds
    except XoaSigmaError as e:
        if errors == "raise":
            raise e
        if errors == "warn":
            xoa_warn("Error while decoding sigma coords: {}. Skipping...".format(e))
        return ds

    # Loop on locations
    from . import sigma as sigma_module

    sigma_type = None
    hsingle = not isinstance(hlocs, (list, tuple))
    if hsingle:
        hlocs = [hlocs]
    for vloc, vterms in all_terms.items():

        for hloc in hlocs:

            terms = vterms if hsingle else vterms[hloc]
            hloc = cfspecs.sglocator.parse_loc_arg(hloc)
            sigma_type = terms.pop("type")

            # Args: keys are cf names, values ara dataarrays
            kwargs = {}
            for vname, cf_name in terms.items():
                if cf_name in HORIZONTAL_TERMS:
                    vname = cfspecs.sglocator.format_attr("name", vname, hloc)
                kwargs[cf_name] = ds[vname]

            # Compute depth/altitude/pressure
            func = getattr(sigma_module, sigma_type)
            cache = {}
            kwargs['cache'] = cache
            height = func(**kwargs)

            # Format
            loc = (hloc or "") + (vloc or "")
            height = cfspecs.sglocator.format_dataarray(height, loc)

            # Transpose if approriate and set as coordinate
            transposed = False
            as_coord = False
            for da in ds.data_vars.values():
                if set(height.dims) <= set(da.dims):
                    if not transposed:
                        height = xcoords.transpose(height, da, "compat")
                        transposed = True
                    ds[da.name] = da.assign_coords({height.name: height})
                    as_coord = True

            # Make sure it is in the dataset
            if not as_coord:
                ds.coords[height.name] = height

    if sigma_type:
        ds.encoding["cf_sigma_type"] = sigma_type
    return ds



#!/usr/bin/env python
# coding: utf-8
"""
Interpolate a meridional section of CROCO outputs to regular depths
===================================================================

This notebook, we show:

* how to compute the depths from s-coordinates,
* how to easily find the name of variables and coordinates,
* how to interpolate a 3D field with varying depths to regular depths.
"""

# %%
# Initialisations
# ---------------
#
# Import needed modules.

import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
import cmocean
import xoa
from xoa.regrid import regrid1d

xr.set_options(display_style="text")

# %%
# Register the :meth:`xarray.Dataset.decode_sigma` callable accessor.

xoa.register_accessors(decode_sigma=True)

# %%
# The :ref:`xoa <accessors>` accessor is also registered by default, and give access
# to most of the fonctionalities of the other accessors.

# %%
# Read the model
# --------------
# This sample is a meridional extraction of a full 3D CROCO output.

ds = xoa.open_data_sample("croco.south-africa.meridional.nc")
print(ds)

# %%
# Compute depths from s-coordinates
# ---------------------------------
#
# Decode the dataset according to the CF conventions:
#
# 1. Find sigma terms
# 2. Compute depths
# 3. Assign depths as coordinates
#
# Note that the :meth:`xarray.Dataset.decode_sigma` callable accessor
# calls the :func:`xoa.sigma.decode_cf_sigma` function.

ds = ds.decode_sigma()
print(ds.depth)

# %%
# Find coordinate names from CF conventions
# -----------------------------------------
#
# The `depth` was assigned as coordinates at the previous stage.
# We use the :ref:`xoa <accessors.dataset>` accessor to easily access the temperature, latitude and depth arrays.
# The default configuration exposes shortcuts for some variables and coordinates
# as shown in :cfsec:`accessors`.

temp = ds.xoa.temp.squeeze()
lat_name = temp.xoa.lat.name

# %%
# Interpolate at regular depths
# -----------------------------
#
# We interpolate the temperature array from irregular to regular depths.
#
# Let's create the output depths.

depth = xr.DataArray(np.linspace(ds.depth.min(), ds.depth.max(), 100),
                     name="depth", dims="depth")

# %%
# Let's interpolate the temperature.

tempz = regrid1d(temp, depth)

# %%
# Plots
# -----
#
# Make a basic comparison plots.

fig, axs = plt.subplots(ncols=2, sharex=True, sharey=True, figsize=(10, 4))
kw = dict(levels=np.arange(0, 23))
temp.plot.contourf(lat_name, "depth", cmap="cmo.thermal", ax=axs[0], **kw)
temp.plot.contour(lat_name, "depth", colors='w', linewidths=.3, ax=axs[0], **kw)
tempz.plot.contourf(lat_name, "depth", cmap="cmo.thermal", ax=axs[1], **kw)
tempz.plot.contour(lat_name, "depth", colors='w', linewidths=.3, ax=axs[1], **kw);


# %%
# Et voilà!



# -*- coding: utf-8 -*-
"""
Test the :mod:`xoa.regrid` module
"""

import numpy as np
import xarray as xr
import pytest

from xoa import regrid
from test_interp import get_grid2locs_coords, vfunc, get_interp1d_data


def test_regrid_regrid1d():

    # Get some data
    xxi, yyi, vari, xxo, yyo = get_interp1d_data()

    # Some inits
    nz1 = xxo.shape[1]
    nx, nz0 = xxi.shape
    nt = 2
    dep0 = xr.DataArray(yyi[0], dims='nz', name='nz')
    dep1 = xr.DataArray(yyo[0], dims='nk', name='nk')
    lon = xr.DataArray(xxi[:, 0], dims='lon')
    time = xr.DataArray(np.arange(nt, dtype='d'), dims='time')

    # data:3d, coord_in:1d, coord_out:1d
    da_in = xr.DataArray(
        np.resize(vari, (nt, nx, nz0)),
        name="banana",
        dims=('time', 'lon', 'nz'),
        coords=(time, lon, dep0),
        attrs={'long_name': 'Big banana'})
    da_out = regrid.regrid1d(da_in, dep1, method="linear")
    assert da_out.dims == ("time", "lon", "nk")
    assert da_out.shape == (nt, nx, nz1)
    assert not np.isnan(da_out).all()
    assert da_out.min() >= da_in.min()
    assert da_out.max() <= da_in.max()

    # data: 3d, coord_in: 2d, coord_out: 2d
    depth_in = xr.DataArray(yyi, dims=("lon", "nz"),)
    depth_out = xr.DataArray(
        yyo, dims=("lon", "nk"),
        attrs={'standard_name': 'ocean_layer_depth'})
    del da_in["nz"]
    da_in = da_in.assign_coords(
        {"time": time, "lon": lon, "depth": depth_in})
    da_out = regrid.regrid1d(da_in, depth_out, method="linear")
    assert da_out.dims == ('time', "lon", "nk")
    assert "depth" in da_out.coords
    assert not np.isnan(da_out).all()

    # same but transposed
    da_in = da_in.transpose("time", "nz", "lon")
    da_out = regrid.regrid1d(da_in, depth_out, method="linear")
    assert da_out.dims == ('time', "nk", "lon")
    assert not np.isnan(da_out).all()


@pytest.mark.parametrize(
    "mode,expected", [
        ["no", [np.nan, 1, np.nan]],
        ["bottom", [1, 1, np.nan]],
        ["below", [1, 1, np.nan]],
        [-1, [1, 1, np.nan]],
        ['top', [np.nan, 1, 1]],
        ['both', [1, 1, 1]],
        ]
    )
def test_regrid_extrap1d(mode, expected):
    nz, ny, nx = 4, 3, 5
    zi = np.linspace(3, 10, nz)
    zi = xr.DataArray(np.arange(nz), dims="z")
    vi = xr.DataArray(np.ones((nz, ny, nx)), dims=('z', 'y', 'x'), coords={"z": zi})
    vi[:, 0] = np.nan
    vi[:, -1] = np.nan
    vi.attrs["long_name"] = "Long name"
    vi.name = "toto"
    vo = regrid.extrap1d(vi, "y", mode)
    np.testing.assert_allclose(vo.values[0, :, 0], expected)
    assert vo.name == vi.name
    assert vo.attrs == vi.attrs
    assert 'z' in vo.coords


def test_regrid_grid2loc():

    np.random.seed(0)

    # Multi-dimensional generic coordinates
    nex = 4
    nexz = 2
    nxi = 7
    nyi = 6
    nzi = 5
    nti = 4
    no = 10
    xxi, yyi, zzi, tti, xo, yo, to, zo = get_grid2locs_coords(
        nex=nex, nexz=nexz, nxi=nxi, nyi=nyi, nzi=nzi, nti=nti, no=no)
    ttidt = tti.astype("m8[us]") + np.datetime64("1950-01-01")
    todt = to.astype("m8[us]") + np.datetime64("1950-01-01")
    todt = xr.DataArray(todt, dims='time')
    xo = xr.DataArray(xo, dims="time")
    yo = xr.DataArray(yo, dims="time")
    zo = xr.DataArray(zo, dims="time")
    loc = xr.Dataset(
        coords={"time": todt, "depth": zo, "lat": yo, "lon": xo})

    # Pure 1D axes
    xi = xr.DataArray(xxi[0, 0, 0, :], dims='lon')
    yi = xr.DataArray(yyi[0, 0, :, 0], dims='lat')
    zi = xr.DataArray(zzi[0, 0, :, 0, 0], dims='depth')
    ti = xr.DataArray(ttidt[:, 0, 0, 0], dims='time')
    mi = xr.DataArray(np.arange(nex), dims='member')
    vi = vfunc(tti, zzi, yyi, xxi)
    vi = xr.DataArray(
        np.resize(vi, (nex, )+vi.shape[1:]),
        dims=('member', 'time', 'depth', 'lat', 'lon'),
        coords={"member": mi, "time": ti, "depth": zi, "lat": yi, "lon": xi},
        attrs={'long_name': "Long name"})
    vo_truth = np.array(vfunc(to, zo.values, yo.values, xo.values))
    vo_interp = regrid.grid2loc(vi, loc)
    assert vo_interp.shape == (nex, no)
    assert vo_interp.dims == ("member", "time")
    assert "time" in vo_interp.coords
    assert "lon" in vo_interp.coords
    assert "member" in vo_interp.coords
    vo_truth[np.isnan(vo_interp[0].values)] = np.nan
    np.testing.assert_almost_equal(vo_interp[0], vo_truth)
    assert "long_name" in vo_interp.attrs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Regridding utilities
"""
# Copyright 2020-2021 Shom
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import xarray as xr

from .__init__ import XoaError
from . import misc
from . import cf as xcf
from . import coords as xcoords
from . import grid as xgrid
from . import interp


class XoaRegridError(XoaError):
    pass


class regrid1d_methods(misc.IntEnumChoices, metaclass=misc.DefaultEnumMeta):
    """Supported :func:`regrid1d` methods"""

    #: Linear iterpolation (default)
    linear = 1
    #: Linear iterpolation (default)
    interp = 1  # compat
    #: Nearest iterpolation
    nearest = 0
    #: Cubic iterpolation
    cubic = 2
    #: Hermitian iterpolation
    hermit = 3
    #: Hermitian iterpolation
    hermitian = 3
    #: Cell-averaging or conservative regridding
    cellave = -1
    #: Cell-averaging or conservative regridding
    cellerr = -2


class extrap_modes(misc.IntEnumChoices, metaclass=misc.DefaultEnumMeta):
    """Supported extrapolation modes"""

    #: No extrapolation (default)
    no = 0
    #: No extrapolation (default)
    none = 0
    #: No extrapolation (default)
    false = 0
    #: Toward the top (after)
    top = 1
    #: Toward the top (after)
    above = 1
    #: Toward the top (after)
    after = 1
    #: Toward the bottom (before)
    bottom = -1
    #: Toward the bottom (before)
    below = -1
    #: Both below and above
    both = 2
    #: Both below and above
    all = 2
    #: Both below and above
    yes = 2
    #: Both below and above
    true = 2


def _wrapper1d_(vari, *args, func_name, **kwargs):
    """To make sure arrays have a 2D shape

    Output array is reshaped back accordindly
    """
    # To 2D
    eshape = vari.shape[:-1]
    args = [np.reshape(arr, (-1, arr.shape[-1])) for arr in (vari,) + args]
    args = [np.asarray(arr) for arr in args]

    # Call
    func = getattr(interp, func_name)
    varo = func(*args, **kwargs)

    # From 2D
    return varo.reshape(eshape + varo.shape[-1:])


def regrid1d(
    da,
    coord,
    method=None,
    dim=None,
    coord_in_name=None,
    edges=None,
    conserv=False,
    extrap="no",
    bias=0.0,
    tension=0.0,
    dask='allowed',
):
    """Regrid along a single dimension

    The input and output coordinates may vary along other dimensions,
    which useful for instance for vertical interpolation in coastal
    ocean models.
    Since it uses func:`xarray.apply_ufunc`, it support dask features.
    The core computation is performed by the numba-accelerated routines
    of :mod:`xoa.interp`.

    Parameters
    ----------
    da: xarray.DataArray
        Array to interpolate
    coord: xarray.DataArray
        Output coordinate
    method: str, int
        Regridding method as one of the following:
        {regrid1d_methods.rst_with_links}
    dim:  str, tuple(str), None
        Dimension on which to operate. If a string, it is expected to
        be the same dimension for both input and output coordinates.
        Else, provide a two-element tuple: ``(dim_in, dim_out)``.
        It is inferred by default from output coordinate et input data array.
    coord_in_name: str, None
        Name of the input coordinate array, which must be known of ``da``.
        It is inferred from the input dara array and dimension name
        by default.
    edges: dict, None
        Grid edge coordinates along the interpolation dimension,
        for the conservative regridding.
        When not provided, edges are computed with :func:`xoa.grid.get_edges`.
        Keys are `"in"` and/or `"out"` and values are arrays with the same shape as
        coordinates except along the interpolation dimension on which 1 is added.
    conserv: bool
        Use conservative regridding when using ``cellave`` method.
    extrap: str, int
        Extrapolation mode as one of the following:
        {extrap_modes.rst_with_links}
    dask: str

    Returns
    -------
    xarray.DataArray
        Regridded array with ``coord`` as new coordinate array.

    See also
    --------
    xoa.interp.nearest1d
    xoa.interp.linear2d
    xoa.interp.cubic2d
    xoa.interp.hermit1d
    xoa.interp.cellave1d
    xoa.interp.extrap1d
    xarray.apply_ufunc
    """
    # Get the working dimensions
    if not isinstance(dim, (tuple, list)):
        dim = (dim, dim)
    dim_in, dim_out = dim
    cfspecs_in = xcf.get_cf_specs(da)
    cfspecs_out = xcf.get_cf_specs(coord)
    # - dim out
    if dim_out is None:  # get dim_out from coord_out
        dim_dict = cfspecs_out.search_dim(coord, errors="raise")
        dim_out = dim_dict["dim"]
        dim_type = dim_dict["type"]
    else:  # dim_out is provided
        dim_type = cfspecs_out.coords.get_dim_type(dim_out, coord)
    # - dim in
    if dim_in is None:
        if dim_type:
            dim_in = cfspecs_in.coords.search_dim(da, dim_type, errors="raise")
        else:
            dim_in = dim_out  # be cafeful, dim1 must be in input!
    assert dim_in in da.dims
    assert dim_out in coord.dims

    # Input coordinate
    if coord_in_name:
        assert coord_in_name in da.coords, 'Invalid coordinate'
        coord_in = da.coords['coord_in_name']
    else:
        coord_in = cfspecs_in.search_coord_from_dim(da, dim_in, errors="raise")
        coord_in_name = coord_in.name

    # Coordinate arguments
    output_sizes = {dim_out: coord.sizes[dim_out]}
    input_core_dims = [[dim_in]]
    method = regrid1d_methods[method]
    coord_out = coord
    exclude_dims = {dim_in, dim_out}
    if int(method) < 0:
        idimin = coord_in.get_axis_num(dim_in)
        idimout = coord.get_axis_num(dim_out)
        if edges and "in" in edges:
            coord_in = edges["in"]
        else:
            coord_in = xgrid.get_edges_1d(coord_in, axis=dim_in)
        if edges and "out" in edges:
            coord = edges["out"]
        else:
            coord = xgrid.get_edges_1d(coord, axis=dim_out)
        namein = coord_in.dims[idimin]
        nameout = coord.dims[idimout]
        input_core_dims.extend([[namein], [nameout]])
        exclude_dims = {dim_in, dim_out, namein, nameout}
    else:
        exclude_dims = {dim_in, dim_out}
        input_core_dims.extend([[dim_in], [dim_out]])
    output_core_dims = [[dim_out]]

    # Interpolation function name and arguments
    func_name = str(method) + "1d"
    if method == regrid1d_methods.cellerr and not (coord_in.ndim == coord.ndim == 1):
        raise XoaRegridError(
            "cellerr regrid method works only with 1D input and output coordinates"
        )
    # func = getattr(interp, func_name)
    extrap = str(extrap_modes[extrap])
    func_kwargs = {"func_name": func_name, "extrap": extrap}
    if method == "hermit":
        func_kwargs.update(bias=bias, tension=tension)

    # Apply
    da_out = xr.apply_ufunc(
        _wrapper1d_,
        da,
        coord_in,
        coord,
        join="override",
        kwargs=func_kwargs,
        input_core_dims=input_core_dims,
        output_core_dims=output_core_dims,
        exclude_dims=exclude_dims,
        dask_gufunc_kwargs={"output_sizes": output_sizes},
        dask=dask,
    )

    # Transpose
    dims = list(da.dims)
    dims[dims.index(dim_in)] = dim_out
    da_out = da_out.transpose(..., *dims, missing_dims="ignore")

    # Add output coordinates
    coord_out_name = coord_out.name if coord_out.name else coord_in.name
    da_out = da_out.assign_coords({coord_out_name: coord_out})
    da_out.name = da.name
    da_out.attrs = da.attrs

    return da_out


regrid1d.__doc__ = regrid1d.__doc__.format(**locals())


def extrap1d(da, dim, mode, **kwargs):
    """Extrapolate along a single dimension


    Parameters
    ----------
    da: xarray.DataArray
        Array to interpolate
    dim:  str
        Dimension on which to operate.
    mode: str, int
        Extrapolation mode as one of the following:
        {extrap_modes.rst_with_links}
    kwargs: dict
        Extra arguments are passed to :func:`xarray.apply_ufunc`

    Returns
    -------
    xarray.DataArray
        Extrapolated array.

    See also
    --------
    xoa.interp.extrap1d
    xarray.apply_ufunc
    """
    da_out = xr.apply_ufunc(
        _wrapper1d_,
        da,
        join="override",
        kwargs={"func_name": "extrap1d", "mode": str(extrap_modes[mode])},
        input_core_dims=[[dim]],
        output_core_dims=[[dim]],
        exclude_dims={dim},
        dask_gufunc_kwargs={"output_sizes": da.sizes},
        **kwargs,
    )
    da_out = da_out.transpose(*da.dims)
    da_out = da_out.assign_coords(da.coords)
    da_out.attrs.update(da.attrs)
    da_out.encoding.update(da.encoding)
    return da_out


extrap1d.__doc__ = extrap1d.__doc__.format(**locals())


def grid2loc(da, loc, compat="warn"):
    """Interpolate a gridded data array to random locations

    ``da`` and ``loc`` must comply with CF conventions.

    Parameters
    ----------
    da: xarray.DataArray
        A data array with at least an horizontal rectilinear or
        a curvilinear grid.
    loc: xarray.Dataset, xarray.DataArray, pandas.DataFrame
        A dataset or data array with coordinates as 1d arrays
        that share the same dimension.
        For example, such dataset may be initialized as follows::

            loc = xr.Dataset(coords={
                'lon': ('npts', [5, 6]),
                'lat': ('npts', [4, 5]),
                'depth': ('npts',  [-10, -20])
                })

    compat: {"ignore", "warn"}
        In case a requested coordinate is not found in the input dataset.

    Return
    ------
    xarray.dataArray
        The interpolated data array.

    See also
    --------
    xoa.interp.grid2locs
    xoa.interp.grid2relloc
    xoa.interp.grid2rellocs
    xoa.interp.cell2relloc
    """

    # Get coordinates
    if hasattr(loc, "to_xarray"):
        loc = loc.to_xarray()
    # - horizontal
    order = "yx"
    lons = xcoords.get_lon(loc)
    lats = xcoords.get_lat(loc)
    xo = lons.values
    yo = lats.values
    # - vertical
    deps = xcoords.get_vertical(loc, errors="ignore")
    if deps is not None:
        gdep = xcoords.get_vertical(da, errors=compat)
        if gdep is not None:
            order = "z" + order
    # - temporal
    times = xcoords.get_time(loc, errors="ignore")
    if times is not None:
        gtime = xcoords.get_time(da, errors=compat)
        if gtime is not None:
            order = "t" + order

    # Transpose following the tzyx order
    glon = xcoords.get_lon(da)  # before to_rect
    glat = xcoords.get_lat(da)  # before to_rect
    dims_in = set(glon.dims).union(glat.dims)
    da_tmp = xgrid.to_rect(da)
    da_tmp = xcoords.reorder(da_tmp, order)

    # To numpy with singletons
    # - data
    vi = da_tmp.data
    for axis_type, axis in (("z", -3), ("t", -4)):
        if axis_type not in order:
            vi = np.expand_dims(vi, axis)
    vi = vi.reshape((-1,) + vi.shape[-4:])
    # - xy
    glon = xcoords.get_lon(da_tmp)  # after to_rect
    glat = xcoords.get_lat(da_tmp)  # after to_rect
    xi = glon.values
    yi = glat.values
    coords_out = [lons, lats]
    if xi.ndim == 1:
        xi = xi.reshape(1, -1)
    if yi.ndim == 1:
        yi = yi.reshape(-1, 1)
    # - z
    if "z" in order:
        gdep_order = xcoords.get_order(da_tmp[gdep.name])
        dims_in.update(gdep.dims)
        zi = da_tmp[gdep.name].values
        for axis_type, axis in (("x", -1), ("y", -2), ("t", -4)):
            if axis_type not in gdep_order:
                zi = np.expand_dims(zi, axis)
        zo = deps.data
        coords_out.append(deps)
    else:
        zi = np.zeros((1, 1, 1, 1))
        zo = np.zeros_like(lons.values)
    zi = zi.reshape((-1,) + zi.shape[-4:])
    # - t
    if "t" in order:
        # numeric times
        ti = (gtime.values - np.datetime64("1950-01-01", "us")) / np.timedelta64(1, "us")
        to = (times.values - np.datetime64("1950-01-01", "us")) / np.timedelta64(1, "us")
        dims_in.update(gtime.dims)
        coords_out.append(times)
    else:
        ti = np.zeros(1)
        to = np.zeros(lons.shape)

    # Interpolate
    vo = interp.grid2locs(xi, yi, zi, ti, vi, xo, yo, zo, to)

    # As data array
    dims_out = [dim for dim in da.dims if dim not in dims_in]
    sizes_out = [size for dim, size in da.sizes.items() if dim in dims_out]
    dims_out.extend(loc.dims)
    sizes_out.append(lons.shape[-1])
    coords_out = coords_out + xcoords.get_coords_compat_with_dims(da, exclude_dims=dims_in)
    da_out = xr.DataArray(
        vo.reshape(sizes_out),
        dims=dims_out,
        coords=dict((coord.name, coord) for coord in coords_out),
        attrs=da.attrs,
        name=da.name,
    )

    # Transpose
    da_out = xcoords.transpose(da_out, da.dims, mode="compat")

    return da_out



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Miscellaneaous low level utilities
"""
# Copyright 2020-2021 Shom
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import types
from enum import IntEnum, EnumMeta

import numpy as np

from .__init__ import XoaError


class XEnumMeta(EnumMeta):
    """Exented version ofnum meta-class

    This version supports:

    - :meth:`__contains__` (``in``) with strings
    - a better :meth:`__str__` (``str()``) method
    - a :meth:`get_rst` methods and :attr`str` and
      :attr:`rst_with_links` properties.

    Example
    -------
    .. ipython:: python

        @suppress
        from xoa.misc import XEnumMeta
        from enum import IntEnum

        class regrid_methods(IntEnum, metaclass=XEnumMeta):
            linear = 1
            bilinear = 1
            nearest = 0
            cellave = -1

        regrid_methods
        'linear' in regrid_methods
        'xxx' in regrid_methods
        1 in regrid_methods
        str(regrid_methods)
        regrid_methods.get_rst(with_links=True, link_module="xoa.tutu")
        regrid_methods.rst

    """
    default = None

    # def __call__(cls, value=None, *args, **kwargs):
    #     if value is None:
    #         return next(iter(cls))
    #     return super().__call__(value, *args, **kwargs)

    def __getitem__(cls, name):
        if not isinstance(name, str):
            return cls(name)
        return cls._member_map_[name]

    def __contains__(cls, value):
        if isinstance(value, str):
            return value in list(cls._member_map_.keys())
        return super().__call__(value)

    def _get_groups_(cls):
        groups = {}
        for name, number in cls._member_map_.items():
            groups.setdefault(number, []).append(name)
        return groups

    def _get_choices_(cls, es=""):
        choices = []
        for number, names in cls._get_groups_().items():
            cc = [f"{number}"] + [f'"{name}"' for name in names]
            choices.append(es + "|".join(cc) + es)
        return choices

    def __str__(cls):
        return ", ".join(cls._get_choices_())

    def __repr__(cls):
        return f"{cls.__name__}: {cls}"

    def get_rst(cls, with_links=False, link_module=None):
        if with_links:
            choices = []
            prefix = ((link_module+".") if link_module else "")
            prefix += cls.__name__ + "."
            for number, names in cls._get_groups_().items():
                number = int(number)
                attr = names[0]
                choice = "|".join([f"{number:d}"] +
                                  [f'"{name}"' for name in names])
                choices.append(f":attr:`{choice}<{prefix}{attr}>`")
            return ", ".join(choices)
        else:
            return ", ".join(cls._get_choices_("``"))

    @property
    def rst(cls):
        return cls.get_rst()

    @property
    def rst_with_links(cls):
        return cls.get_rst(with_links=True)


class DefaultEnumMeta(XEnumMeta):
    """Enum meta-class that support default value and None

    When the item is not provided or equal to ``None``, the first
    declared item is returned

    Inspired from: https://stackoverflow.com/questions/44867597/is-there-a-way-to-specify-a-default-value-for-python-enums

    Example
    -------
    .. ipython:: python

        @suppress
        from xoa.misc import DefaultEnumMeta
        from enum import IntEnum

        class regrid_methods(IntEnum, metaclass=DefaultEnumMeta):
            linear = 1
            bilinear = 1
            nearest = 0
            cellave = -1

        regrid_methods() # default method
        regrid_methods(None) # default method
        regrid_methods(1)
        regrid_methods[None] # default method
        regrid_methods['linear']
        regrid_methods['cellave']

    """
    default = None

    def __call__(cls, value=None, *args, **kwargs):
        if value is None:
            return next(iter(cls))
        return super().__call__(value, *args, **kwargs)

    def __getitem__(cls, name):
        if name is None:
            return next(iter(cls))
        if not isinstance(name, str):
            return cls(name)
        return cls._member_map_[name]


class IntEnumChoices(IntEnum):

    def __str__(self):
        return self.name


class Choices(object):
    """Choice management for a function or method parameter

    Parameters
    ----------
    choices: dict, list
        Allowed choices. When a dict, values are considered as the
        description of choices.
    case_insensitive: bool
        Wether the treatment of string type choice should be case
        sensitive or not.
    parameter: None, str
        Parameter name, which defaults to the lower case class name
    description: str
        Short description of the parameter.
    """

    def __init__(self, choices, case_insensitive=True, parameter=None,
                 description="Choices"):
        self._ci = case_insensitive
        self._docs = {}
        if isinstance(choices, dict):
            self._choices = []
            for value, doc in choices.items():
                value = self._reformat_value_(value)
                self._docs[value] = doc
                self._choices.append(value)
        else:
            self._choices = [self._reformat_value_(value) for value in choices]
        self._parameter = (self.__class__.__name__.lower()
                           if parameter is None else parameter)
        self._description = description

    @property
    def choices(self):
        return self._choices

    def _reformat_value_(self, value):
        if self._ci and isinstance(value, str):
            return value.lower()
        return value

    def __getitem__(self, choice):
        choice = self._reformat_value_(choice)
        if choice not in self._choices:
            desc = self._description if self._description else 'choice'
            raise XoaError(f"Invalid choice for \"{desc}\": {choice}. "
                           f"Please choose one of: {self}")
        return choice

    def __str__(self):
        return ", ".join(self._choices)

    def to_docstring(self, indent=4):
        """Convert to numpy-like docstring

        Parameters
        ----------
        indent: str, int
            Base indentation. Integers are multiplied by a space char.

        Return
        ------
        str
            Docstring of this parameter
        """
        indent = (" " * indent) if isinstance(indent, int) else indent
        pindent = indent + 4 * " "
        types = "{" + ", ".join([repr(c) for c in self.choices]) + "}"
        rst = f"{self._parameter}: {types}\n"
        if self._description:
            rst += f"{pindent}{self._description}\n"
        if self._docs:
            rst += "\n"
            for choice, doc in self._docs.items():
                rst += f"{pindent}- ``{choice}``: {doc}\n"
            # rst += "\n"
        return rst

    def format_function_docstring(self, func):
        func.__doc__ = func.__doc__.format(
            **{self._parameter: self.to_docstring(4)})
        return func

    def format_method_docstring(self, func):
        func.__doc__ = func.__doc__.format(
            **{self._parameter: self.to_docstring(8)})
        return func


ERRORS = Choices({"ignore": "silently ignore",
                  "warn": "emit a warning",
                  "raise": "raise an exception"},
                 parameter="errors",
                 description="In case of errors"
                 )


def get_axis_slices(ndim, axis, **kwargs):
    """Get standard slices for an axis of a ndim array

    Parameters
    ----------
    ndim:
        The number of dimensions. It can also be
        a tuple (like an array shape) or an array.
    axis:
        Index of the axis.

    Return
    ------
    A dictionary of tuples of slices. All tuples have a
        length of ndim, and can be used has a slice for the array
        (see exedges =ample).

        - ``"all"``: Select everything.
        - ``"first"``/``"last"``: First and last.
        - ``"firstp1"``: Second element.
        - ``"firstp2"``: Third element.
        - ``"lastm1"``: Element before the last one.
        - ``"lastm2"``: Second element before the last one.
        - ``"firsts"``: All but the last.
        - ``"lasts"``: All but the first.
        - ``"firstsm1"``: All but the last two.
        - ``"lastsp1"``: All but the first two.
        - ``"mid"``: All but the first and last.

    Example
    -------
    .. ipython:: python

        @suppress
        import numpy as np, pprint
        @suppress
        from xoa.misc import get_axis_slices

        var = np.arange(2*5*4).reshape(2, 5, 4)
        pprint.pprint(get_axis_slices(var, axis=1))
    """
    if not isinstance(ndim, int):
        ndim = np.ndim(ndim)
    sel = [slice(None)]*ndim
    selmid = list(sel)
    selmid[axis] = slice(1, -1)
    selmid = tuple(selmid)
    sellasts = list(sel)
    sellasts[axis] = slice(1, None)
    sellasts = tuple(sellasts)
    selfirsts = list(sel)
    selfirsts[axis] = slice(0, -1)
    selfirsts = tuple(selfirsts)
    sellastsp1 = list(sel)
    sellastsp1[axis] = slice(2, None)
    sellastsp1 = tuple(sellastsp1)
    selfirstsm1 = list(sel)
    selfirstsm1[axis] = slice(0, -2)
    selfirstsm1 = tuple(selfirstsm1)
    sellast = list(sel)
    sellast[axis] = -1
    sellast = tuple(sellast)
    selfirst = list(sel)
    selfirst[axis] = 0
    selfirst = tuple(selfirst)
    sellastm1 = list(sel)
    sellastm1[axis] = -2
    sellastm1 = tuple(sellastm1)
    sellastm2 = list(sel)
    sellastm2[axis] = -3
    sellastm2 = tuple(sellastm2)
    selfirstp1 = list(sel)
    selfirstp1[axis] = 1
    selfirstp1 = tuple(selfirstp1)
    selfirstp2 = list(sel)
    selfirstp2[axis] = 2
    selfirstp2 = tuple(selfirstp2)
    if kwargs:
        for key, val in kwargs.items():
            if isinstance(val, (list, tuple)):
                val = slice(*val)
            ksel = list(sel)
            ksel[axis] = val
            kwargs[key] = ksel
    return dict(all=sel, mid=selmid, lasts=sellasts, firsts=selfirsts,
                lastsp1=sellastsp1, firstsm1=selfirstsm1,
                last=sellast, first=selfirst, lastm1=sellastm1,
                firstp1=selfirstp1,
                lastm2=sellastm2, firstp2=selfirstp2, **kwargs)


def is_iterable(obj, nostr=True, nogen=True):
    """Check if an object is iterable or not

    Parameters
    ----------
    obj:
        Object to check

    Return
    ------
    bool
    """

    if not nogen and isinstance(obj, types.GeneratorType):
        return True
    if not (hasattr(obj, "__len__") and callable(obj.__len__)):
        return False
    if nostr:
        return not isinstance(obj, str)
    return True


def dict_check_defaults(dd, **defaults):
    """Check that a dictionary has some default values

    Parameters
    ----------
    dd: dict
        Dictionary to check
    **defs: dict
        Dictionary of default values

    Example
    -------
    .. ipython:: python

        @suppress
        from xoa.misc import dict_check_defaults

        dd = dict(color='blue')
        dict_check_defaults(dd, color='red', size=10)
    """
    if defaults is None:
        defaults = {}
    for item in defaults.items():
        dd.setdefault(*item)
    return dd


def dict_filter(
    kwargs,
    filters,
    defaults=None,
    copy=False,
    short=False,
    keep=False,
    **kwadd,
):
    """Filter out kwargs (typically extra calling keywords)

    Parameters
    ----------
    kwargs:
        Dictionnary to filter.
    filters:
        Single or list of prefixes.
    defaults:
        dictionnary of default values for output fictionnary.
    copy:
        Simply copy items, do not remove them from kwargs.
    short:
        Allow prefixes to not end with ``"_"``.
    keep:
        Keep prefix filter in output keys.

    Example
    -------
    .. ipython:: python

        @suppress
        from xoa.misc import dict_filter
        kwargs = {'basemap':'f', 'basemap_fillcontinents':True, 'quiet':False,'basemap_plot':False}
        dict_filter(kwargs,'basemap', defaults=dict(drawcoastlines=True,plot=True), good=True)
        kwargs

    Return
    ------
    dict
    """

    if isinstance(filters, str):
        filters = [filters]
    if copy:
        kwread = kwargs.get
    else:
        kwread = kwargs.pop

    # Set initial items
    kwout = {}
    for filter_ in filters:
        if not filter_.endswith("_") and filter_ in kwargs:
            if isinstance(kwargs[filter_], dict):
                kwout.update(kwread(filter_))
            else:
                kwout[filter_] = kwread(filter_)
        if not short and not filter_.endswith("_"):
            filter_ += "_"
        for att, val in list(kwargs.items()):
            if att.startswith(filter_) and att != filter_:
                if keep:
                    kwout[att] = kwread(att)
                else:
                    kwout[att[len(filter_):]] = kwread(att)

    # Add some items
    kwout.update(kwadd)

    # Set some default values
    if defaults is not None:
        for att, val in defaults.items():
            kwout.setdefault(att, val)
    return kwout


def dict_merge(*dd, mergesubdicts=True, mergelists=False, mergetuples=False,
               uniquify=False, skipnones=True, overwriteempty=False, cls=None,
               **kwargs):
    """Merge dictionaries

    First dictionaries have priority over next

    Parameters
    ----------
    dd:
        Argument are interpreted as dictionary to merge.
        Those who are not dictionaries are skipped.
    mergesubdicts: optional
        Also merge dictionary items
        (like in a tree).
    mergetuples: optional
        Also merge tuple items.
    mergelists: optional
        Also merge list items.
    uniquify: optional
        Uniquify lists and tuples.
    skipnones: optional
        Skip Nones.
    overwriteempty: optional
        Overwrite value that does are not True when converted to bool.
    cls: optional
        Class to use. Default to the first class found in arguments
        that is not a :class:`dict`, else defaults to :class:`dict`.

    Example
    -------
    .. ipython:: python

        @suppress
        from xoa.misc import dict_merge
        d1 = dict(a=3, b=5, e=[1, 2])
        d2 = dict(a=5, c=7, e=[3, 4])
        print(dict_merge(d1, d2, mergelists=True))

    """
    # Options
    dd = [_f for _f in dd if _f]

    # Get the class
    if cls is None:
        cls = dict
        for d in dd:
            if d.__class__ is not dict:
                cls = d.__class__
                break

    # Init
    from configobj import Section, ConfigObj

    if cls is Section:
        for d in dd:
            if isinstance(d, Section):
                break
        else:
            raise XoaError("Can't initialise Section for merging")
        outd = Section(d.parent, d.depth, d.main, name=d.name)
    else:
        outd = cls()
    kwargs.update(
        mergesubdicts=mergesubdicts,
        mergelists=mergelists,
        mergetuples=mergetuples,
        uniquify=uniquify,
        skipnones=skipnones,
        overwriteempty=overwriteempty,
        cls=cls)

    # Loop
    for d in dd:
        if not isinstance(d, dict):
            continue

        # Content
        for key, val in d.items():
            if skipnones and val is None:
                continue

            # Not set so we set
            if key not in outd or (overwriteempty and is_empty(outd[key])):
                outd[key] = val

            # Merge subdict
            elif (
                mergesubdicts
                and isinstance(outd[key], dict)
                and isinstance(val, dict)
            ):
                outd[key] = dict_merge(outd[key], val, **kwargs)

            # Merge lists
            elif (
                mergelists
                and isinstance(outd[key], list)
                and isinstance(val, list)
            ):
                outd[key] += val
                if uniquify:
                    outd[key] = gunique(list(outd[key]))

            # Merge tuples
            elif (
                mergetuples
                and isinstance(outd[key], tuple)
                and isinstance(val, tuple)
            ):
                outd[key] += val
                if uniquify:
                    outd[key] = tuple(gunique(outd[key]))

    # Comments for ConfigObj instances
    if cls is ConfigObj:
        if not outd.initial_comment and hasattr(d, "initial_comment"):
            outd.initial_comment = d.initial_comment
        if not outd.final_comment and hasattr(d, "final_comment"):
            outd.final_comment = d.final_comment
        if hasattr(d, "inline_comments") and d.inline_comments:
            outd.inline_comments = dict_merge(
                outd.inline_comments, d.inline_comments, overwriteempty=True
            )

    return outd


def is_empty(x):
    """Check if empty"""
    if isinstance(x, bool):
        return False
    try:
        return not bool(x)
    except Exception:
        return False


def match_string(ss, checks, ignorecase=True, transform=None):
    """Check that a string verify a check list that consists of
    a list of either strings or callables

    Parameters
    ----------
    ss: str
    checks: str, callable, list of {str or callable}
    ignorecase: bool
    transform: callable

    Example
    -------
    .. ipython:: python

        @suppress
        from xoa.misc import match_string
        import re
        match_string('sst', 'sst')
        match_string('sst', [re.compile(r'ss.$').match])

    Return
    ------
    True
    """
    # Nothing
    if not ss or not checks:
        return False

    # Setup
    ss = ss.strip()
    if ignorecase:
        ss = ss.lower()
    if not is_iterable(checks, nogen=False):
        checks = [checks]
    checks = [x for x in checks if x is not None]

    # Callables
    sss = []
    for check in checks:
        if callable(transform) and not callable(check):
            check = transform(check)
        if callable(check) and check(ss):
            return True
        if isinstance(check, str):
            sss.append(check)

    # Strings
    sss = [s.strip() for s in sss]
    if ignorecase:
        sss = [s.lower() for s in sss]
    return ss in sss


def match_attrs(obj, checks, ignorecase=True, transform=None):
    """Check that at least one of the attributes matches check list

    Parameters
    ----------
    obj: object
    checks: dict
        A dictionary of (attribute name, checklist), checklist being an
        iterable as accepted by :func:`match_string`.
    """
    if obj is None or checks is None:
        return False
    for attname, attchecks in checks.items():
        if hasattr(obj, attname) and match_string(
            getattr(obj, attname),
            attchecks,
            ignorecase=ignorecase,
            transform=transform,
        ):
            return True
    return False


def gunique(seq):
    """Create a generator that yields unique item whlist presrving the order

    Parameters
    ----------
    seq: sequence

    Yields
    ------
    item

    Example
    -------
    .. ipython:: python

        @suppress
        from xoa.misc import gunique
        print(list(gunique([1, 6, 1, 8])))
    """
    seen = set()
    seen_add = seen.add
    for x in seq:
        if not (x in seen or seen_add(x)):
            yield x


class ArgList(object):
    """Utility to always manage arguments as list and return results as input

    Examples
    --------
    .. ipython:: python

        @suppress
        from xoa.misc import ArgList

        # Scalar
        a = 'a'
        al = ArgList(a)
        al.get() # input for function as tuple
        al.put(['aa']) # output as input

        # Iterable
        a = ('a','b')
        al = ArgList(a)
        al.get()
        al.put(['aa'])

    """

    def __init__(self, argsi):
        self.single = not isinstance(argsi, list)
        self.argsi = argsi

    def get(self):
        return [self.argsi] if self.single else self.argsi

    def put(self, argso):
        so = not isinstance(argso, list)
        if (so and self.single) or (not so and not self.single):
            return argso
        if so and not self.single:
            return [argso]
        return argso[0]


class ArgTuple(object):
    """Utility to always manage arguments as tuple and return results as input

    Examples
    --------
    .. ipython:: python

        @suppress
        from xoa.misc import ArgTuple

        # Scalar
        a = 'a'
        al = ArgTuple(a)
        al.get() # input for function as tuple
        al.put(('aa',)) # output as input

        # Iterable
        a = ('a','b')
        al = ArgTuple(a)
        al.get()
        al.put(('aa',))

    """

    def __init__(self, argsi):
        self.single = not isinstance(argsi, tuple)
        self.argsi = argsi

    def get(self):
        return (self.argsi,) if self.single else self.argsi

    def put(self, argso):
        so = not isinstance(argso, tuple)
        if (so and self.single) or (not so and not self.single):
            return argso
        if so and not self.single:
            return (argso,)
        return argso[0]
